"""
Provides interactive tools based on matplotlib Qt interactive elements.
This are generally primitive one offs that are useful for accomplishing
something quick. As examples:

1. `pca_explorer` lets you interactively examine a PCA decomposition or
   other decomposition supported by `arpes.analysis.decomposition`
2. `pick_points`, `pick_rectangles` allows selecting many individual points
    or regions from a piece of data, useful to isolate locations to do
    further analysis.
3. `kspace_tool` allows interactively setting coordinate offset for
    angle-to-momentum conversion.
4. `fit_initializer` allows for seeding an XPS curve fit.

All of these return a "context" object which can be used to get information from the current
session (i.e. the selected points or regions, or modified data).
If you forget to save this context, you can recover it as the most recent context
is saved at `arpes.config.CONFIG` under the key "CURRENT_CONTEXT".

There are also primitives for building interactive tools in matplotlib. Such as
DataArrayView, which provides an interactive and updatable plot view from an
xarray.DataArray instance.

In the future, it would be nice to get higher quality interactive tools, as
we start to run into the limits of these ones. But between this and `qt_tool`
we are doing fine for now.
"""

import itertools
from functools import wraps

import matplotlib
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np

from matplotlib.path import Path
from matplotlib.widgets import (Button, LassoSelector, RectangleSelector, Slider, SpanSelector, TextBox)

import arpes.config
from arpes.fits import LorentzianModel, broadcast_model
from arpes.plotting.utils import fancy_labels, imshow_arr, invisible_axes
from arpes.utilities import normalize_to_spectrum
from arpes.utilities.conversion import convert_to_kspace

__all__ = ('pick_rectangles', 'pick_points', 'pca_explorer', 'kspace_tool', 'fit_initializer',)


class SelectFromCollection:
    """
    Select indices from a matplotlib collection using `LassoSelector`.

    Modified from https://matplotlib.org/gallery/widgets/lasso_selector_demo_sgskip.html

    Selected indices are saved in the `ind` attribute. This tool fades out the
    points that are not part of the selection (i.e., reduces their alpha
    values). If your collection has alpha < 1, this tool will permanently
    alter the alpha values.

    Note that this tool selects collection objects based on their *origins*
    (i.e., `offsets`).
    """

    def __init__(self, ax, collection, alpha_other=0.3, on_select=None):
        self.canvas = ax.figure.canvas
        self.collection = collection
        self.alpha_other = alpha_other

        self.xys = collection.get_offsets()
        self.n_pts = len(self.xys)
        self._on_select = on_select

        # Ensure that we have separate colors for each object
        self.facecolors = collection.get_facecolors()
        if not self.facecolors:
            raise ValueError('Collection must have a facecolor')

        if len(self.facecolors) == 1:
            self.facecolors = np.tile(self.facecolors, (self.n_pts, 1))

        self.lasso = LassoSelector(ax, onselect=self.onselect)
        self.ind = []

    def onselect(self, verts):
        try:
            path = Path(verts)
            self.ind = np.nonzero(path.contains_points(self.xys))[0]
            self.facecolors[:, -1] = self.alpha_other
            self.facecolors[self.ind, -1] = 1
            self.collection.set_facecolors(self.facecolors)
            self.canvas.draw_idle()

            if self._on_select is not None:
                self._on_select(self.ind)
        except Exception:
            pass

    def disconnect(self):
        self.lasso.disconnect_events()
        self.facecolors[:, -1] = 1
        self.collection.set_facecolors(self.facecolors)
        self.canvas.draw_idle()


def popout(plotting_function):
    """
    Sets and subsequently unsets the matplotlib backend for one function call, to allow use of
    'widgets' in Jupyter inline use.

    :param plotting_function:
    :return:
    """

    @wraps(plotting_function)
    def wrapped(*args, **kwargs):
        from IPython import get_ipython

        ipython = get_ipython()
        ipython.magic('matplotlib qt')

        return plotting_function(*args, **kwargs)

        # ideally, cleanup, but this closes the plot, necessary but redundant looking import
        # look into an on close event for matplotlib
        #ipython.magic('matplotlib inline')
        #from matplotlib import pyplot as plt

    return wrapped


class DataArrayView:
    """
    Offers support for 1D and 2D DataArrays with masks, selection tools, and a simpler interface
    than the matplotlib primitives.

    Look some more into holoviews for different features. https://github.com/pyviz/holoviews/pull/1214
    """
    def __init__(self, ax, data=None, ax_kwargs=None, mask_kwargs=None, transpose_mask=False, auto_autoscale=True):
        self.ax = ax
        self._initialized = False
        self._data = None
        self._mask = None
        self.n_dims = None
        self.ax_kwargs = ax_kwargs or dict()
        self._axis_image = None
        self._mask_image = None
        self._mask_cmap = None
        self._transpose_mask = transpose_mask
        self._selector = None
        self._inner_on_select = None
        self.auto_autoscale = auto_autoscale
        self.mask_kwargs = mask_kwargs

        if data is not None:
            self.data = data

    def handle_select(self, event_click=None, event_release=None):
        dims = self.data.dims

        if self.n_dims == 2:
            x1, y1 = event_click.xdata, event_click.ydata
            x2, y2 = event_release.xdata, event_release.ydata

            x1, x2 = min(x1, x2), max(x1, x2)
            y1, y2 = min(y1, y2), max(y1, y2)

            region = dict([[dims[1], slice(x1, x2)], [dims[0], slice(y1, y2)]])
        else:
            x1, x2 = event_click, event_release
            x1, x2 = min(x1, x2), max(x1, x2)

            region = dict([[self.data.dims[0], slice(x1, x2)]])

        self._inner_on_select(region)

    def attach_selector(self, on_select):
        # data should already have been set
        assert self.n_dims is not None

        self._inner_on_select = on_select

        if self.n_dims == 1:
            self._selector = SpanSelector(
                self.ax, self.handle_select, 'horizontal',
                useblit=True, rectprops=dict(alpha=0.35, facecolor='red'),
            )
        else:
            self._selector = RectangleSelector(
                self.ax, self.handle_select, drawtype='box',
                rectprops=dict(fill=False, edgecolor='black', linewidth=2),
                lineprops=dict(linewidth=2, color='black'),
            )

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, new_data):
        if self._initialized:
            self._data = new_data
        else:
            self._data = new_data
            self._initialized = True
            self.n_dims = len(new_data.dims)
            if self.n_dims == 2:
                self._axis_image = imshow_arr(self._data, ax=self.ax, **self.ax_kwargs)[1]
                fancy_labels(self.ax)
            else:
                self.ax_kwargs.pop('cmap', None)
                x, y = self.data.coords[self.data.dims[0]].values, self.data.values
                self._axis_image = self.ax.plot(x, y, **self.ax_kwargs)
                self.ax.set_xlabel(self.data.dims[0])
                cs = self.data.coords[self.data.dims[0]].values
                self.ax.set_xlim([np.min(cs), np.max(cs)])
                fancy_labels(self.ax)

        if self.n_dims == 2:
            x, y = self._data.coords[self._data.dims[0]].values, self._data.coords[self._data.dims[1]].values
            extent = [y[0], y[-1], x[0], x[-1]]
            self._axis_image.set_extent(extent)
            self._axis_image.set_data(self._data.values)
        else:
            color = self.ax.lines[0].get_color()
            self.ax.lines.remove(self.ax.lines[0])
            x, y = self.data.coords[self.data.dims[0]].values, self.data.values
            l, h = np.min(y), np.max(y)
            self._axis_image = self.ax.plot(x, y, c=color, **self.ax_kwargs)
            self.ax.set_ylim([l - 0.1 * (h - l), h + 0.1 * (h - l)])

        if self.auto_autoscale:
            self.autoscale()

    @property
    def mask_cmap(self):
        if self._mask_cmap is None:
            self._mask_cmap = matplotlib.cm.get_cmap(self.mask_kwargs.pop('cmap', 'Reds'))
            self._mask_cmap.set_bad('k', alpha=0)

        return self._mask_cmap

    @property
    def mask(self):
        return self._mask

    @mask.setter
    def mask(self, new_mask):
        if np.array(new_mask).shape != self.data.values.shape:
            # should be indices then
            mask = np.zeros(self.data.values.shape, dtype=bool)
            np.ravel(mask)[new_mask] = True
            new_mask = mask

        self._mask = new_mask

        for_mask = np.ma.masked_where(np.logical_not(self._mask), self.data.values * 0 + 1)
        if self.n_dims == 2 and self._transpose_mask:
            for_mask = for_mask.T

        if self.n_dims == 2:
            if self._mask_image is None:
                self._mask_image = self.ax.imshow(
                    for_mask.T, cmap=self.mask_cmap, interpolation='none', vmax=1, vmin=0,
                    origin='lower', extent=self._axis_image.get_extent(),
                    aspect=self.ax.get_aspect(), **self.mask_kwargs
                )
            else:
                self._mask_image.set_data(for_mask.T)
        else:
            if self._mask_image is not None:
                self.ax.collections.remove(self._mask_image)

            x = self.data.coords[self.data.dims[0]].values
            low, high = self.ax.get_ylim()
            calc_cmap = self.mask_cmap
            self._mask_image = self.ax.fill_between(x, low, for_mask * high, color=calc_cmap(1.), **self.mask_kwargs)

    def autoscale(self):
        if self.n_dims == 2:
            self._axis_image.autoscale()
        else:
            pass


@popout
def fit_initializer(data, peak_type=LorentzianModel, **kwargs):
    ctx = {}
    gs = gridspec.GridSpec(2, 2)
    ax_initial = plt.subplot(gs[0, 0])
    ax_fitted = plt.subplot(gs[0, 1])
    ax_other = plt.subplot(gs[1, 0])
    ax_test = plt.subplot(gs[1, 1])

    invisible_axes(ax_other)

    prefixes = 'abcdefghijklmnopqrstuvwxyz'
    model_settings = []
    model_defs = []
    fitted_individual_models = []
    for_fit = data.expand_dims('fit_dim')
    for_fit.coords['fit_dim'] = np.array([0])

    data_view = DataArrayView(ax_initial)
    residual_view = DataArrayView(ax_fitted, ax_kwargs=dict(linestyle=':', color='orange'))
    fitted_view = DataArrayView(ax_fitted, ax_kwargs=dict(color='red'))
    initial_fit_view = DataArrayView(ax_fitted, ax_kwargs=dict(linestyle='--', color='blue'))

    def compute_parameters():
        renamed = [{'{}_{}'.format(prefix, k): v for k, v in m_setting.items()} for m_setting, prefix in zip(model_settings, prefixes)]
        return dict(itertools.chain(*[list(d.items()) for d in renamed]))

    def on_add_new_peak(selection):
        amplitude = data.sel(**selection).mean().item()

        selection = selection[data.dims[0]]
        center = (selection.start + selection.stop) / 2
        sigma = (selection.stop - selection.start)

        model_settings.append({'center': {'value': center, 'min': center - sigma, 'max': center + sigma},
                               'sigma': {'value': sigma}, 'amplitude': {'min': 0, 'value': amplitude}})
        model_defs.append(LorentzianModel)

        if model_defs:
            results = broadcast_model(model_defs, for_fit, 'fit_dim', params=compute_parameters())
            result = results.results[0].item()

            if result is not None:
                # residual
                for_residual = data.copy(deep=True)
                for_residual.values = result.residual
                residual_view.data = for_residual

                # fit_result
                for_best_fit = data.copy(deep=True)
                for_best_fit.values = result.best_fit
                fitted_view.data = for_best_fit

                # initial_fit_result
                for_initial_fit = data.copy(deep=True)
                for_initial_fit.values = result.init_fit
                initial_fit_view.data = for_initial_fit

                ax_fitted.set_ylim(ax_initial.get_ylim())


    data_view.data = data
    data_view.attach_selector(on_select=on_add_new_peak)
    ctx['data'] = data

    def on_copy_settings(event):
        try:
            import pyperclip
            import pprint
            pyperclip.copy(pprint.pformat(compute_parameters()))
        except ImportError:
            pass
        finally:
            import pprint
            print(pprint.pformat(compute_parameters()))

    copy_settings_button = Button(ax_test, 'Copy Settings')
    copy_settings_button.on_clicked(on_copy_settings)
    ctx['button'] = copy_settings_button
    return ctx


@popout
def pca_explorer(pca, data, component_dim='components', initial_values=None, transpose_mask=False, **kwargs):
    if initial_values is None:
        initial_values = [0, 1]

    pca_dims = list(pca.dims)
    pca_dims.remove(component_dim)
    other_dims = [d for d in data.dims if d not in pca_dims]

    context = {
        'selected_components': initial_values,
        'selected_indices': [],
        'sum_data': None,
        'map_data': None,
        'selector': None,
        'integration_region': {},
    }
    arpes.config.CONFIG['CURRENT_CONTEXT'] = context

    def compute_for_scatter():
        for_scatter = pca.copy(deep=True).isel(**dict([[component_dim, context['selected_components']]]))
        for_scatter = for_scatter.S.transpose_to_back(component_dim)

        size = data.mean(other_dims).stack(pca_dims=pca_dims).values
        norm = np.expand_dims(np.linalg.norm(pca.values, axis=(0,)), axis=-1)

        return (for_scatter / norm).stack(pca_dims=pca_dims), 5 * size / np.mean(size)

    # ===== Set up axes ======
    gs = gridspec.GridSpec(2, 2)
    ax_components = plt.subplot(gs[0, 0])
    ax_sum_selected = plt.subplot(gs[0, 1])
    ax_map = plt.subplot(gs[1, 0])

    gs_widget = gridspec.GridSpecFromSubplotSpec(3, 1, subplot_spec=gs[1, 1])
    ax_widget_1 = plt.subplot(gs_widget[0, 0])
    ax_widget_2 = plt.subplot(gs_widget[1, 0])
    ax_widget_3 = plt.subplot(gs_widget[2, 0])

    selected_view = DataArrayView(ax_sum_selected, ax_kwargs=dict(cmap='viridis'))
    map_view = DataArrayView(ax_map, ax_kwargs=dict(cmap='Greys'),
                             mask_kwargs=dict(cmap='Reds', alpha=0.35), transpose_mask=transpose_mask)

    def update_from_selection(ind):
        # Calculate the new data
        if ind is None or not ind:
            context['selected_indices'] = []
            context['sum_data'] = data.stack(pca_dims=pca_dims).sum('pca_dims')
        else:
            context['selected_indices'] = ind
            context['sum_data'] = data.stack(pca_dims=pca_dims).isel(pca_dims=ind).sum('pca_dims')

        if context['integration_region']:
            data_sel = data.sel(**context['integration_region']).sum(other_dims)
        else:
            data_sel = data.sum(other_dims)

        # Update all views
        map_view.data = data_sel
        map_view.mask = ind
        selected_view.data = context['sum_data']

    def set_axes(component_x, component_y):
        ax_components.clear()
        context['selected_components'] = [component_x, component_y]
        for_scatter, size = compute_for_scatter()
        pts = ax_components.scatter(for_scatter.values[0], for_scatter.values[1], s=size)

        if context['selector'] is not None:
            context['selector'].disconnect()

        context['selector'] = SelectFromCollection(ax_components, pts, on_select=update_from_selection)
        ax_components.set_xlabel('$e_' + str(component_x) + '$')
        ax_components.set_ylabel('$e_' + str(component_y) + '$')
        update_from_selection([])

    def on_change_axes(event):
        try:
            val_x = int(context['axis_X_input'].text)
            val_y = int(context['axis_Y_input'].text)

            def clamp(x, low, high):
                if low <= x < high:
                    return x
                if x < low:
                    return low
                return high

            maximum = len(pca.coords[component_dim].values) - 1

            val_x, val_y = clamp(val_x, 0, maximum), clamp(val_y, 0, maximum)

            assert val_x != val_y

            set_axes(val_x, val_y)
        except Exception:
            pass

    context['axis_button'] = Button(ax_widget_1, 'Change Decomp Axes')
    context['axis_button'].on_clicked(on_change_axes)
    context['axis_X_input'] = TextBox(ax_widget_2, 'Axis X:', initial=str(initial_values[0]))
    context['axis_Y_input'] = TextBox(ax_widget_3, 'Axis Y:', initial=str(initial_values[1]))

    def on_select_summed(region):
        context['integration_region'] = region
        update_from_selection(context['selected_indices'])

    set_axes(*initial_values)
    selected_view.attach_selector(on_select_summed)

    plt.tight_layout()
    return context


@popout
def kspace_tool(data, **kwargs):
    original_data = data
    data = normalize_to_spectrum(data)
    if len(data.dims) > 2:
        data = data.sel(eV=slice(-0.05, 0.05)).sum('eV', keep_attrs=True)
        data.coords['eV'] = 0

    if 'eV' in data.dims:
        data = data.S.transpose_to_front('eV')

    data = data.copy(deep=True)

    ctx = {
        'original_data': original_data,
        'data': data,
        'widgets': []
    }
    arpes.config.CONFIG['CURRENT_CONTEXT'] = ctx
    gs = gridspec.GridSpec(4, 3)
    ax_initial = plt.subplot(gs[0:2, 0:2])
    ax_converted = plt.subplot(gs[2:, 0:2])

    n_widget_axes = 8
    gs_widget = gridspec.GridSpecFromSubplotSpec(n_widget_axes, 1, subplot_spec=gs[:, 2])

    widget_axes = [plt.subplot(gs_widget[i, 0]) for i in range(n_widget_axes)]
    [invisible_axes(a) for a in widget_axes[:-2]]

    skip_dims = {'x', 'X', 'y', 'y', 'z', 'Z', 'T'}
    for dim in skip_dims:
        if dim in data.dims:
            raise ValueError('Please provide data without the {} dimension'.format(dim))

    convert_dims = ['theta', 'beta', 'phi', 'psi']
    if 'eV' not in data.dims:
        convert_dims += ['chi']
    if 'hv' in data.dims:
        convert_dims += ['hv']

    ang_range = (-45 * np.pi / 180, 45 * np.pi / 180, 0.01)
    default_ranges = {
        'eV': [-0.05, 0.05, 0.001],
        'hv': [-20, 20, 0.5],
    }

    sliders = {}

    def update_kspace_plot(_):
        for name, slider in sliders.items():
            data.attrs['{}_offset'.format(name)] = slider.val

        converted_view.data = convert_to_kspace(data)

    axes = iter(widget_axes)
    for convert_dim in convert_dims:
        widget_ax = next(axes)
        low, high, delta = default_ranges.get(convert_dim, ang_range)
        init = data.S.lookup_offset(convert_dim)
        sliders[convert_dim] = Slider(widget_ax, convert_dim, init + low, init + high, valinit=init, valstep=delta)
        sliders[convert_dim].on_changed(update_kspace_plot)

    def compute_offsets():
        return {k: v.val for k, v in sliders.items()}

    def on_copy_settings(event):
        try:
            import pyperclip
            import pprint
            pyperclip.copy(pprint.pformat(compute_offsets()))
        except ImportError:
            pass
        finally:
            import pprint
            print(pprint.pformat(compute_offsets()))

    def apply_offsets(event):
        for name, offset in compute_offsets().items():
            print(name, offset)
            original_data.attrs['{}_offset'.format(name)] = offset
            try:
                for s in original_data.S.spectra:
                    s.attrs['{}_offset'.format(name)] = offset
            except AttributeError:
                pass

    ctx['widgets'].append(sliders)

    copy_settings_button = Button(widget_axes[-1], 'Copy Offsets')
    apply_settings_button = Button(widget_axes[-2], 'Apply Offsets')
    copy_settings_button.on_clicked(on_copy_settings)
    apply_settings_button.on_clicked(apply_offsets)
    ctx['widgets'].append(copy_settings_button)
    ctx['widgets'].append(apply_settings_button)

    data_view = DataArrayView(ax_initial)
    converted_view = DataArrayView(ax_converted)

    data_view.data = data
    update_kspace_plot(None)

    plt.tight_layout()

    return ctx


@popout
def pick_rectangles(data, **kwargs):
    ctx = {'points': [], 'rect_next': False}
    arpes.config.CONFIG['CURRENT_CONTEXT'] = ctx

    rects = []

    fig = plt.figure()
    data.S.plot(**kwargs)
    ax = fig.gca()

    def onclick(event):
        ctx['points'].append([event.xdata, event.ydata])
        if ctx['rect_next']:
            p1, p2 = ctx['points'][-2], ctx['points'][-1]
            p1[0], p2[0] = min(p1[0], p2[0]), max(p1[0], p2[0])
            p1[1], p2[1] = min(p1[1], p2[1]), max(p1[1], p2[1])

            rects.append([p1, p2])
            rect = plt.Rectangle((p1[0], p1[1],), p2[0] - p1[0], p2[1] - p1[1],
                                 edgecolor='red', linewidth=2, fill=False)
            ax.add_patch(rect)

        ctx['rect_next'] = not ctx['rect_next']
        plt.draw()

    _ = plt.connect('button_press_event', onclick)

    return rects


@popout
def pick_gamma(data, **kwargs):
    fig = plt.figure()
    data.S.plot(**kwargs)

    ax = fig.gca()
    dims = data.dims

    def onclick(event):

        data.attrs['symmetry_points'] = {
            'G': {}
        }

        print(event.x, event.xdata, event.y, event.ydata)

        for dim, value in zip(dims, [event.ydata, event.xdata]):
            if dim == 'eV':
                continue

            data.attrs['symmetry_points']['G'][dim] = value

        plt.draw()

    _ = plt.connect('button_press_event', onclick)

    return data


@popout
def pick_points(data, **kwargs):
    ctx = {'points': []}
    arpes.config.CONFIG['CURRENT_CONTEXT'] = ctx

    fig = plt.figure()
    data.S.plot(**kwargs)
    ax = fig.gca()

    x0, y0 = ax.transAxes.transform((0, 0))  # lower left in pixels
    x1, y1 = ax.transAxes.transform((1, 1))  # upper right in pixes
    dx = x1 - x0
    dy = y1 - y0
    maxd = max(dx, dy)
    xlim, ylim = ax.get_xlim(), ax.get_ylim()

    width = .03 * maxd / dx * (xlim[1] - xlim[0])
    height = .03 * maxd / dy * (ylim[1] - ylim[0])

    def onclick(event):
        ctx['points'].append([event.xdata, event.ydata])

        circ = matplotlib.patches.Ellipse((event.xdata, event.ydata,), width, height, color='red')
        ax.add_patch(circ)

        plt.draw()

    _ = plt.connect('button_press_event', onclick)

    return ctx['points']



"""
The core of this module is `broadcast_model` which is a serious workhorse in PyARPES for
analyses based on curve fitting. This allows simple multidimensional curve fitting by
iterative fitting across one or many axes. Currently basic strategies are implemented,
but in the future we would like to provide:

1. Passing xr.DataArray values to parameter guesses and bounds, which can be interpolated/selected
   to allow changing conditions throughout the curve fitting session.
2. A strategy allowing retries with initial guess taken from the previous fit. This is similar
   to some adaptive curve fitting routines that have been proposed in the literature.
"""

import functools
import operator
import warnings
from string import ascii_lowercase

import lmfit
import numpy as np
from tqdm import tqdm_notebook

import arpes.fits.fit_models
import typing
import xarray as xr
from arpes.provenance import update_provenance
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum

__all__ = ('broadcast_model',)


TypeIterable = typing.Union[typing.List[type], typing.Tuple[type]]



def parse_model(model):
    """
    Takes a model string and turns it into a tokenized version.

    1. ModelClass -> ModelClass
    2. [ModelClass] -> [ModelClass]
    3. str -> [<ModelClass, operator as string>]

    i.e.

    A + (B + C) * D -> [A, '(', B, '+', C, ')', '*', D]

    :param model:
    :return:
    """
    if not isinstance(model, str):
        return model

    pad_all = ['+', '-', '*', '/', '(', ')']

    for pad in pad_all:
        model = model.replace(pad, ' {} '.format(pad))

    special = set(pad_all)

    def read_token(token):
        if token in special:
            return token
        try:
            token = float(token)
            return token
        except ValueError:
            try:
                return arpes.fits.fit_models.__dict__[token]
            except KeyError:
                raise ValueError('Could not find model: {}'.format(token))

    return [read_token(token) for token in model.split()]


def _parens_to_nested(items):
    """
    Turns a flat list with parentheses tokens into a nested list
    :param items:
    :return:
    """

    parens = [(token, idx,) for idx, token in enumerate(items) if isinstance(token, str) and token in '()']
    if parens:
        first_idx, last_idx = parens[0][1], parens[-1][1]
        if parens[0][0] != '(' or parens[-1][0] != ')':
            raise ValueError('Parentheses do not match!')

        return items[0:first_idx] + [_parens_to_nested(items[first_idx + 1:last_idx])] + items[last_idx + 1:]
    else:
        return items


def reduce_model_with_operators(model):
    if isinstance(model, tuple):
        return model[0](prefix='{}_'.format(model[1]), nan_policy='omit')

    if isinstance(model, list) and len(model) == 1:
        return reduce_model_with_operators(model[0])

    left, op, right = model[0], model[1], model[2:]
    left, right = reduce_model_with_operators(left), reduce_model_with_operators(right)

    if op == '+':
        return left + right
    elif op == '*':
        return left * right
    elif op == '-':
        return left - right
    elif op == '/':
        return left / right


def compile_model(model, params=None, prefixes=None):
    """
    Takes a model sequence, i.e. a Model class, a list of such classes, or a list
    of such classes with operators and instantiates an appropriate model.
    :param model:
    :return:
    """
    if params is None:
        params = {}

    prefix_compile = '{}'
    if prefixes is None:
        prefixes = ascii_lowercase
        prefix_compile = '{}_'

    try:
        if issubclass(model, lmfit.Model):
            return model()
    except TypeError:
        pass

    if isinstance(model, (list, tuple)) and all([isinstance(token, type) for token in model]):
        models = [m(prefix=prefix_compile.format(prefixes[i]), nan_policy='omit') for i, m in enumerate(model)]
        if isinstance(params, (list, tuple)):
            for cs, m in zip(params, models):
                for name, params_for_name in cs.items():
                    m.set_param_hint(name, **params_for_name)

        built = functools.reduce(operator.add, models)
    else:
        warnings.warn('Beware of equal operator precedence.')
        prefix = iter(prefixes)
        model = [m if isinstance(m, str) else (m, next(prefix)) for m in model]
        built = reduce_model_with_operators(_parens_to_nested(model))

    return built


def unwrap_params(params, iter_coordinate):
    """
    Inspects parameters to see if any are array like and extracts the appropriate value to use for the current
    iteration point.
    :param params:
    :param iter_coordinate:
    :return:
    """
    def transform_or_walk(v):
        if isinstance(v, dict):
            return unwrap_params(v, iter_coordinate)

        if isinstance(v, xr.DataArray):
            return v.sel(**iter_coordinate, method='nearest').item()

        return v

    return {k: transform_or_walk(v) for k, v in params.items()}


@update_provenance('Broadcast a curve fit along several dimensions')
def broadcast_model(model_cls: typing.Union[type, TypeIterable],
                    data: DataType, broadcast_dims, params=None, progress=True, dataset=True,
                    weights=None, safe=False, prefixes=None):
    """
    Perform a fit across a number of dimensions. Allows composite models as well as models
    defined and compiled through strings.
    :param model_cls:
    :param data:
    :param broadcast_dims:
    :param params:
    :param progress:
    :param dataset:
    :param weights:
    :param safe:
    :return:
    """
    if params is None:
        params = {}

    if isinstance(broadcast_dims, str):
        broadcast_dims = [broadcast_dims]

    data = normalize_to_spectrum(data)
    cs = {}
    for dim in broadcast_dims:
        cs[dim] = data.coords[dim]

    other_axes = set(data.dims).difference(set(broadcast_dims))
    template = data.sum(list(other_axes))
    template.values = np.ndarray(template.shape, dtype=np.object)

    residual = data.copy(deep=True)
    residual.values = np.zeros(residual.shape)

    model = compile_model(parse_model(model_cls), params=params, prefixes=prefixes)
    if isinstance(params, (list, tuple)):
        params = {}

    new_params = model.make_params()

    n_fits = np.prod(np.array(list(template.S.dshape.values())))
    wrap_progress = lambda x, *args, **kwargs: x
    if progress:
        wrap_progress = tqdm_notebook

    for indices, cut_coords in wrap_progress(template.T.enumerate_iter_coords(), desc='Fitting', total=n_fits):
        current_params = unwrap_params(params, cut_coords)

        cut_data = data.sel(**cut_coords)
        if safe:
            cut_data = cut_data.T.drop_nan()

        weights_for = None
        if weights is not None:
            weights_for = weights.sel(**cut_coords)

        try:
            fit_result = model.guess_fit(cut_data, params=current_params, weights=weights_for)
        except ValueError:
            fit_result = None

        template.loc[cut_coords] = fit_result

        try:
            residual.loc[cut_coords] = fit_result.residual if fit_result is not None else None
        except ValueError as e:
            if not safe:
                raise e

    if dataset:
        return xr.Dataset({
            'results': template,
            'data': data,
            'residual': residual,
            'norm_residual': residual / data,
        }, residual.coords)

    template.attrs['original_data'] = data
    return template



import numpy as np

from arpes.provenance import update_provenance
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum

__all__ = ('remove_incoherent_background',)


@update_provenance('Remove incoherent background from above Fermi level')
def remove_incoherent_background(data: DataType, set_zero=True):
    """
    Sometimes spectra are contaminated by data above the Fermi level for
    various reasons (such as broad core levels from 2nd harmonic light,
    or slow enough electrons in ToF experiments to be counted in subsequent
    pulses).
    :param data:
    :param set_zero:
    :return:
    """
    data = normalize_to_spectrum(data)

    approximate_fermi_energy_level = data.S.find_spectrum_energy_edges().max()

    background = data.sel(eV=slice(approximate_fermi_energy_level + 0.1, None))
    density = background.sum('eV') / (np.logical_not(np.isnan(background)) * 1).sum('eV')
    new = data - density
    if set_zero:
        new.values[new.values < 0] = 0

    return new



import collections
import datetime
import errno
import itertools
import json
import os.path
import warnings
from collections import Counter

import matplotlib
import matplotlib.cm
import matplotlib.offsetbox
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colorbar, colors, gridspec
from matplotlib.lines import Line2D

import xarray as xr
from arpes import VERSION
from arpes.config import CONFIG, FIGURE_PATH
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum

__all__ = (
    # General + IO
    'path_for_plot', 'path_for_holoviews', 'name_for_dim', 'unit_for_dim',
    'savefig', 'AnchoredHScaleBar', 'calculate_aspect_ratio',

    # color related
    'temperature_colormap',
    'polarization_colorbar',
    'temperature_colormap_around',
    'temperature_colorbar',
    'temperature_colorbar_around',

    'generic_colorbarmap',
    'generic_colorbarmap_for_data',
    'colorbarmaps_for_axis',

    # Axis generation
    'dos_axes',

    # matplotlib 'macros'
    'invisible_axes',
    'no_ticks',
    'get_colorbars',
    'remove_colorbars',
    'frame_with',

    'imshow_arr',
    'imshow_mask',
    'lineplot_arr',  # 1D version of imshow_arr
    'plot_arr',  # generic dimension version of imshow_arr, plot_arr

    # insets related
    'inset_cut_locator',
    'swap_xaxis_side',
    'swap_yaxis_side',
    'swap_axis_sides',

    # TeX related
    'quick_tex',

    # Decorating + labeling
    'label_for_colorbar', 'label_for_dim', 'label_for_symmetry_point',
    'sum_annotation',
    'fancy_labels',

    # Data summaries
    'summarize',

    'transform_labels',
)


def swap_xaxis_side(ax):
    ax.xaxis.tick_top()
    ax.xaxis.set_label_position("top")


def swap_yaxis_side(ax):
    ax.yaxis.tick_right()
    ax.yaxis.set_label_position("right")


def swap_axis_sides(ax):
    swap_xaxis_side(ax)
    swap_yaxis_side(ax)


def transform_labels(transform_fn, fig=None, include_titles=True):
    if fig is None:
        fig = plt.gcf()

    axes = list(fig.get_axes())
    for ax in axes:
        try:
            ax.set_xlabel(transform_fn(ax.get_xlabel(), is_title=False))
            ax.set_ylabel(transform_fn(ax.get_xlabel(), is_title=False))
            if include_titles:
                ax.set_title(transform_fn(ax.get_title(), is_title=True))
        except TypeError:
            ax.set_xlabel(transform_fn(ax.get_xlabel()))
            ax.set_ylabel(transform_fn(ax.get_xlabel()))
            if include_titles:
                ax.set_title(transform_fn(ax.get_title()))


def summarize(data: DataType, axes=None):
    data = normalize_to_spectrum(data)

    axes_shapes_for_dims = {
        1: (1, 1),
        2: (1, 1),
        3: (2, 2),  # one extra here
        4: (3, 2),  # corresponds to 4 choose 2 axes
    }

    if axes is None:
        fig, axes = plt.subplots(axes_shapes_for_dims.get(len(data.dims)), figsize=(8, 8))

    flat_axes = axes.ravel()
    combinations = list(itertools.combinations(data.dims, 2))
    for axi, combination in zip(flat_axes, combinations):
        data.sum(combination).plot(ax=axi)
        fancy_labels(axi)

    for i in range(len(combinations), len(flat_axes)):
        flat_axes[i].set_axis_off()

    return axes


def sum_annotation(eV=None, phi=None):
    eV_annotation, phi_annotation = '', ''

    def to_str(bound):
        if bound is None:
            return ''

        return '{:.2f}'.format(bound)

    if eV is not None:
        eV_annotation = '$\\text{E}_{' + to_str(eV.start) + '}^{' + to_str(eV.stop) + '}$'
    if phi is not None:
        phi_annotation = '$\\phi_{' + to_str(phi.start) + '}^{' + to_str(phi.stop) + '}$'

    return eV_annotation + phi_annotation


def frame_with(ax, color='red', linewidth=2):
    for spine in ['left', 'right', 'top', 'bottom']:
        ax.spines[spine].set_color(color)
        ax.spines[spine].set_linewidth(linewidth)


def quick_tex(latex_fragment, ax=None, fontsize=30):
    """
    Sometimes you just need to render some latex and getting a latex session
    running is far too much effort.
    :param latex_fragment:
    :return:
    """

    if ax is None:
        fig, ax = plt.subplots()

    invisible_axes(ax)
    ax.text(0.2, 0.2, latex_fragment, fontsize=fontsize)
    return ax


def lineplot_arr(arr, ax=None, method='plot', mask=None, mask_kwargs=None, **kwargs):
    if mask_kwargs is None:
        mask_kwargs = dict()

    if ax is None:
        _, ax = plt.subplots()

    xs = None
    if arr is not None:
        fn = plt.plot
        if method == 'scatter':
            fn = plt.scatter

        xs = arr.coords[arr.dims[0]].values
        fn(xs, arr.values, **kwargs)

    if mask is not None:
        y_lim = ax.get_ylim()
        if isinstance(mask, list) and isinstance(mask[0], slice):
            for slice_mask in mask:
                ax.fill_betweenx(y_lim, slice_mask.start, slice_mask.stop, **mask_kwargs)
        else:
            raise NotImplementedError()
        ax.set_ylim(y_lim)

    return ax


def plot_arr(arr=None, ax=None, over=None, mask=None, **kwargs):
    to_plot = arr if mask is None else mask
    try:
        n_dims = len(to_plot.dims)
    except AttributeError:
        n_dims = 1

    if n_dims == 2:
        quad = None
        if arr is not None:
            ax, quad = imshow_arr(arr, ax=ax, over=over, **kwargs)
        if mask is not None:
            over = quad if over is None else over
            imshow_mask(mask, ax=ax, over=over, **kwargs)
    if n_dims == 1:
        ax = lineplot_arr(arr, ax=ax, mask=mask, **kwargs)

    return ax


def imshow_mask(mask, ax=None, over=None, cmap=None, **kwargs):
    assert over is not None

    if ax is None:
        ax = plt.gca()

    if cmap is None:
        cmap = 'Reds'

    if isinstance(cmap, str):
        cmap = matplotlib.cm.get_cmap(name=cmap)

    cmap.set_bad('k', alpha=0)

    ax.imshow(mask.values, cmap=cmap, interpolation='none', vmax=1, vmin=0,
              origin='lower', extent=over.get_extent(), aspect=ax.get_aspect(), **kwargs)


def imshow_arr(arr, ax=None, over=None, origin='lower', aspect='auto', **kwargs):
    """
    Similar to plt.imshow but users different default origin, and sets appropriate
    extent on the plotted data.
    :param arr:
    :param ax:
    :return:
    """

    if ax is None:
        fig, ax = plt.subplots()

    x, y = arr.coords[arr.dims[0]].values, arr.coords[arr.dims[1]].values
    extent = [y[0], y[-1], x[0], x[-1]]

    if over is None:
        quad = ax.imshow(arr.values, origin=origin, extent=extent, aspect=aspect, **kwargs)
        ax.grid(False)
        ax.set_xlabel(arr.dims[1])
        ax.set_ylabel(arr.dims[0])
    else:
        quad = ax.imshow(arr.values, extent=over.get_extent(), aspect=ax.get_aspect(), origin=origin, **kwargs)

    return ax, quad


def dos_axes(orientation='horiz', figsize=None, with_cbar=True):
    """
    Orientation option should be 'horiz' or 'vert'

    :param orientation:
    :param figsize:
    :param with_cbar:
    :return:
    """
    if figsize is None:
        figsize = (12, 9,) if orientation == 'vert' else (9, 9)

    fig = plt.figure(figsize=figsize)

    outer_grid = gridspec.GridSpec(4, 4, wspace=0.0, hspace=0.0)

    if orientation == 'horiz':
        fig.subplots_adjust(hspace=0.00)
        gs = gridspec.GridSpec(2, 1, height_ratios=[3, 1])

        ax0 = plt.subplot(gs[0])
        axes = (ax0, plt.subplot(gs[1], sharex=ax0))
        plt.setp(axes[0].get_xticklabels(), visible=False)
    else:
        fig.subplots_adjust(wspace=0.00)
        gs = gridspec.GridSpec(1, 2, width_ratios=[1, 4])

        ax0 = plt.subplot(gs[1])
        axes = (ax0, plt.subplot(gs[0], sharey=ax0))
        plt.setp(axes[0].get_yticklabels(), visible=False)

    return fig, axes


def inset_cut_locator(data, reference_data=None, ax=None, location=None, color=None, **kwargs):
    """
    Plots a reference cut location
    :param data:
    :param reference_data:
    :param ax:
    :param location:
    :param kwargs:
    :return:
    """
    quad = data.plot(ax=ax)
    ax.set_xlabel('')
    ax.set_ylabel('')
    try:
        quad.colorbar.remove()
    except Exception:
        pass

    # add more as necessary
    missing_dim_resolvers = {
        'theta': lambda: reference_data.S.theta,
        'beta': lambda: reference_data.S.beta,
        'phi': lambda: reference_data.S.phi,
    }

    missing_dims = [d for d in data.dims if d not in location]
    missing_values = {d: missing_dim_resolvers[d]() for d in missing_dims}
    ordered_selector = [location.get(d, missing_values.get(d)) for d in data.dims]

    n = 200

    def resolve(name, value):
        if isinstance(value, slice):
            low = value.start
            high = value.stop

            if low is None:
                low = data.coords[name].min().item()
            if high is None:
                high = data.coords[name].max().item()

            return np.linspace(low, high, n)

        return np.ones((n,)) * value

    n_cut_dims = len([d for d in ordered_selector if isinstance(d, (collections.Iterable, slice))])
    ordered_selector = [resolve(d, v) for d, v in zip(data.dims, ordered_selector)]

    if missing_dims:
        assert reference_data is not None
        print(missing_dims)

    if n_cut_dims == 2:
        # a region cut, illustrate with a rect or by suppressing background
        return
    if color is None:
        color = 'red'

    if n_cut_dims == 1:
        # a line cut, illustrate with a line
        ax.plot(*ordered_selector[::-1], color=color, **kwargs)
        pass
    elif n_cut_dims == 0:
        # a single point cut, illustrate with a marker

        pass


def generic_colormap(low, high):
    delta = (high - low)
    low = low - delta / 6
    high = high + delta / 6

    def get_color(value):
        return matplotlib.cm.Blues(float((value - low) / (high - low)))

    return get_color


def phase_angle_colormap(low=0, high=np.pi * 2):
    def get_color(value):
        return matplotlib.cm.twilight_shifted(float((value - low) / (high - low)))

    return get_color


def delay_colormap(low=-1, high=1):
    def get_color(value):
        return matplotlib.cm.coolwarm(float((value - low) / (high - low)))

    return get_color


def temperature_colormap(high=300, low=0, cmap=None):
    if cmap is None:
        cmap = matplotlib.cm.Blues_r

    def get_color(value):
        return cmap(float((value - low) / (high - low)))

    return get_color


def temperature_colormap_around(central, range=50):
    def get_color(value):
        return matplotlib.cm.RdBu_r(float((value - central) / range))

    return get_color


def generic_colorbar(low, high, label='', ax=None, ticks=None, **kwargs):
    extra_kwargs = {
        'orientation': 'horizontal',
        'label': label,
        'ticks': ticks if ticks is not None else [low, high],
    }

    delta = (high - low)
    low = low - delta / 6
    high = high + delta / 6

    extra_kwargs.update(kwargs)
    cb = colorbar.ColorbarBase(
        ax, cmap='Blues', norm=colors.Normalize(vmin=low, vmax=high), **extra_kwargs)

    return cb


def phase_angle_colorbar(high=np.pi * 2, low=0, ax=None, **kwargs):
    extra_kwargs = {
        'orientation': 'horizontal',
        'label': 'Angle',
        'ticks': ['0', r'$\pi$', r'$2\pi$']
    }
    extra_kwargs.update(kwargs)
    cb = colorbar.ColorbarBase(ax, cmap='twilight_shifted', norm=colors.Normalize(vmin=low, vmax=high), **extra_kwargs)
    return cb


def temperature_colorbar(high=300, low=0, ax=None, cmap=None, **kwargs):
    if cmap is None:
        cmap = 'Blues_r'

    extra_kwargs = {
        'orientation': 'horizontal',
        'label': 'Temperature (K)',
        'ticks': [low, high],
    }
    extra_kwargs.update(kwargs)
    cb = colorbar.ColorbarBase(ax, cmap=cmap, norm=colors.Normalize(vmin=low, vmax=high), **extra_kwargs)
    return cb


def delay_colorbar(low=-1, high=1, ax=None, **kwargs):
    # TODO make this nonsequential for use in case where you want to have a long time period after the
    # delay or before
    extra_kwargs = {
        'orientation': 'horizontal',
        'label': 'Probe Pulse Delay (ps)',
        'ticks': [low, 0, high],
    }
    extra_kwargs.update(kwargs)
    cb = colorbar.ColorbarBase(ax, cmap='coolwarm', norm=colors.Normalize(vmin=low, vmax=high), **extra_kwargs)
    return cb


def temperature_colorbar_around(central, range=50, ax=None, **kwargs):
    extra_kwargs = {
        'orientation': 'horizontal',
        'label': 'Temperature (K)',
        'ticks': [central - range, central + range],
    }
    extra_kwargs.update(kwargs)
    cb = colorbar.ColorbarBase(ax, cmap='RdBu_r', norm=colors.Normalize(vmin=central - range, vmax=central + range),
                               **extra_kwargs)
    return cb


colorbarmaps_for_axis = {
    'temp': (temperature_colorbar, temperature_colormap,),
    'delay': (delay_colorbar, delay_colormap,),
    'theta': (phase_angle_colorbar, phase_angle_colormap,),
}


def get_colorbars(fig=None):
    if fig is None:
        fig = plt.gcf()

    colorbars = []
    for ax in fig.axes:
        if ax.get_aspect() == 20:
            colorbars.append(ax)

    return colorbars


def remove_colorbars(fig=None):
    """Removes colorbars from given (or, if no given figure, current) matplotlib figure.

    :param fig (default plt.gcf()):
    """

    # TODO after colorbar removal, plots should be relaxed/rescaled to occupy space previously allocated to colorbars
    # for now, can follow this with plt.tight_layout()
    try:
        if fig is not None:
            for ax in fig.axes:
                if ax.get_aspect() == 20:  # a bit of a hack
                    ax.remove()
        else:
            remove_colorbars(plt.gcf())
    except Exception:
        pass


generic_colorbarmap = (generic_colorbar, generic_colormap,)


def generic_colorbarmap_for_data(data: xr.DataArray, keep_ticks=True, ax=None, **kwargs):
    low, high = data.min().item(), data.max().item()
    ticks = None
    if keep_ticks:
        ticks = data.values
    return (
        generic_colorbar(low=low, high=high, ax=ax, ticks=kwargs.get('ticks', ticks)),
        generic_colormap(low=low, high=high),
    )


def polarization_colorbar(ax=None):
    cb = colorbar.ColorbarBase(ax, cmap='RdBu', norm=colors.Normalize(vmin=-1, vmax=1),
                               orientation='horizontal', label='Polarization', ticks=[-1, 0, 1])
    return cb


def calculate_aspect_ratio(data: DataType):
    data = normalize_to_spectrum(data)

    assert len(data.dims) == 2

    x_extent = np.ptp(data.coords[data.dims[0]].values)
    y_extent = np.ptp(data.coords[data.dims[1]].values)

    return y_extent / x_extent


class AnchoredHScaleBar(matplotlib.offsetbox.AnchoredOffsetbox):
    """
    Modified from `this StackOverflow question <https://stackoverflow.com/questions/43258638/>`_
    as alternate to the one provided through matplotlib

    size: length of bar in data units
    extent : height of bar ends in axes units
    """

    def __init__(self, size=1, extent=0.03, label="", loc=2, ax=None,
                 pad=0.4, borderpad=0.5, ppad=0, sep=2, prop=None,
                 label_color=None,
                 frameon=True, **kwargs):
        if not ax:
            ax = plt.gca()
        trans = ax.get_xaxis_transform()

        size_bar = matplotlib.offsetbox.AuxTransformBox(trans)
        line = Line2D([0, size], [0, 0], **kwargs)
        vline1 = Line2D([0, 0], [-extent / 2., extent / 2.], **kwargs)
        vline2 = Line2D([size, size], [-extent / 2., extent / 2.], **kwargs)
        size_bar.add_artist(line)
        size_bar.add_artist(vline1)
        size_bar.add_artist(vline2)
        txt = matplotlib.offsetbox.TextArea(label, minimumdescent=False, textprops={
            'color': label_color,
        })
        self.vpac = matplotlib.offsetbox.VPacker(children=[size_bar, txt],
                                                 align="center", pad=ppad, sep=sep)
        matplotlib.offsetbox.AnchoredOffsetbox.__init__(self, loc, pad=pad,
                                                        borderpad=borderpad, child=self.vpac, prop=prop,
                                                        frameon=frameon)


def savefig(desired_path, dpi=400, data=None, **kwargs):
    full_path = path_for_plot(desired_path)

    if data is not None:
        assert isinstance(data, (list, tuple, set,))

        def extract(for_data):
            try:
                return for_data.attrs.get('provenance', {})
            except Exception:
                return {}

        provenance_context = {
            'VERSION': VERSION,
            'time': datetime.datetime.now().isoformat(),
            'name': 'savefig',
            'data': [extract(d) for d in data],
        }
        provenance_path = full_path + '.provenance.json'
        with open(provenance_path, 'w') as f:
            json.dump(provenance_context, f, indent=2)

    plt.savefig(full_path, dpi=dpi, **kwargs)


def path_for_plot(desired_path):
    workspace = CONFIG['WORKSPACE']

    if not workspace:
        warnings.warn('Saving locally, no workspace found.')
        return os.path.join(os.getcwd(), desired_path)

    if isinstance(workspace, dict):
        workspace = workspace['name']

    filename = os.path.join(FIGURE_PATH, workspace,
                            datetime.date.today().isoformat(), desired_path)
    if not os.path.exists(os.path.dirname(filename)):
        try:
            os.makedirs(os.path.dirname(filename))
        except OSError as exc:
            if exc.errno != errno.EEXIST:
                raise

    return filename


def path_for_holoviews(desired_path):
    skip_paths = ['.svg', '.png', '.jpeg', '.jpg', '.gif']

    prefix, ext = os.path.splitext(desired_path)

    if ext in skip_paths:
        return prefix

    return prefix + ext


def name_for_dim(dim_name, escaped=True):
    name = {
        'beta': r'$\beta$',
        'theta': r'$\theta$',
        'chi': r'$\chi$',
        'alpha': r'$\alpha$',
        'psi': r'$\psi$',
        'phi': r'$\varphi$',
        'eV': r'$\textnormal{E}$',
        'kx': r'$\textnormal{k}_\textnormal{x}$',
        'ky': r'$\textnormal{k}_\textnormal{y}$',
        'kz': r'$\textnormal{k}_\textnormal{z}$',
        'kp': r'$\textnormal{k}_\textnormal{\parallel}$',
        'hv': r'$h\nu$'
    }.get(dim_name)

    if not escaped:
        name = name.replace('$', '')

    return name


def unit_for_dim(dim_name, escaped=True):
    unit = {
        'theta': r'rad',
        'beta': r'rad',
        'psi': r'rad',
        'chi': r'rad',
        'alpha': r'rad',
        'phi': r'rad',
        'eV': r'eV',
        'kx': r'$\AA^{-1}$',
        'ky': r'$\AA^{-1}$',
        'kz': r'$\AA^{-1}$',
        'kp': r'$\AA^{-1}$',
        'hv': r'eV'
    }.get(dim_name)

    if not escaped:
        unit = unit.replace('$', '')

    return unit


def label_for_colorbar(data):
    if not data.S.is_differentiated:
        return r'Spectrum Intensity (arb).'

    # determine which axis was differentiated
    hist = data.S.history
    records = [h['record'] for h in hist if isinstance(h, dict)]
    if 'curvature' in [r['by'] for r in records]:
        curvature_record = [r for r in records if r['by'] == 'curvature'][0]
        directions = curvature_record['directions']
        return r'Curvature along {} and {}'.format(
            name_for_dim(directions[0]),
            name_for_dim(directions[1])
        )

    derivative_records = [r for r in records if r['by'] == 'dn_along_axis']
    c = Counter(itertools.chain(*[[d['axis']] * d['order'] for d in derivative_records]))

    partial_frag = r''
    if sum(c.values()) > 1:
        partial_frag = r'^' + str(sum(c.values()))

    return r'$\frac{\partial' + partial_frag + r' \textnormal{Int.}}{' + \
           r''.join([r'\partial {}^{}'.format(name_for_dim(item, escaped=False), n)
                     for item, n in c.items()]) + '}$ (arb.)'


def label_for_dim(data=None, dim_name=None, escaped=True):
    raw_dim_names = {
        'theta': r'$\theta$',
        'beta': r'$\beta$',
        'chi': r'$\chi$',
        'alpha': r'$\alpha$',
        'psi': r'$\psi$',
        'phi': r'$\varphi$',
        'eV': r'\textbf{eV}',
        'angle': r'Interp. \textbf{Angle}',
        'kinetic': r'Kinetic Energy (\textbf{eV})',
        'temp': r'\textbf{Temperature}',
        'kp': r'$\textbf{k}_\parallel$',
        'kz': r'$\textbf{k}_\perp$',
        'hv': 'Photon Energy'
    }

    if data is not None:
        if data.S.spectrometer.get('type') == 'hemisphere':
            raw_dim_names['phi'] = r'$\varphi$ (Hemisphere Acceptance)'

    if dim_name in raw_dim_names:
        return raw_dim_names.get(dim_name)

    # Next we will look at the listed symmetry_points to try to infer the appropriate way to display the axis
    try:
        from titlecase import titlecase
    except ImportError:
        warnings.warn('Using alternative titlecase, for better results `pip install titlecase`.')

        def titlecase(s):
            """
            Poor man's titlecase
            :param s:
            :return:
            """
            return s.title()

    return titlecase(dim_name.replace('_', ' '))


def fancy_labels(ax_or_ax_set, data=None):
    """
    Attaches better display axis labels for all axes that can be traversed in the
    passed figure or axes.

    :param ax_or_ax_set:
    :param data:
    :return:
    """
    if isinstance(ax_or_ax_set, (list, tuple, set, np.ndarray)):
        for ax in ax_or_ax_set:
            fancy_labels(ax)
        return

    ax = ax_or_ax_set
    try:
        ax.set_xlabel(label_for_dim(data=data, dim_name=ax.get_xlabel()))
    except Exception:
        pass

    try:
        ax.set_ylabel(label_for_dim(data=data, dim_name=ax.get_ylabel()))
    except Exception:
        pass


def label_for_symmetry_point(point_name):
    proper_names = {
        'G': r'$\Gamma$',
        'X': r'X',
        'Y': r'Y',
    }
    return proper_names.get(point_name, point_name)


class CoincidentLinesPlot():
    """
    Helper to allow drawing lines at the same location. Will draw n lines offset so that their
    center appears at the data center, and the lines will end up nonoverlapping.

    Only works for straight lines

    Technique adapted from `StackOverflow
    <https://stackoverflow.com/questions/19394505/matplotlib-expand-the-line-with-specified-width-in-data-unit>`_.
    """

    linewidth = 3

    def __init__(self, **kwargs):
        self.ax = kwargs.pop('ax', plt.gca())
        self.fig = kwargs.pop('fig', plt.gcf())
        self.extra_kwargs = kwargs
        self.ppd = 72. / self.fig.dpi
        self.has_drawn = False

        self.events = {
            'resize_event': self.ax.figure.canvas.mpl_connect('resize_event', self._resize),
            'motion_notify_event': self.ax.figure.canvas.mpl_connect('motion_notify_event', self._resize),
            'button_release_event': self.ax.figure.canvas.mpl_connect('button_release_event', self._resize),
        }
        self.handles = []
        self.lines = []  # saved args and kwargs for plotting, does not verify coincidence

    def add_line(self, *args, **kwargs):
        assert not self.has_drawn
        self.lines.append((args, kwargs,))

    def draw(self):
        self.has_drawn = True

        offset_in_data_units = self.data_units_per_pixel * self.linewidth
        self.offsets = [offset_in_data_units * (o - (len(self.lines) - 1) / 2) for o in range(len(self.lines))]

        for offset, (line_args, line_kwargs) in zip(self.offsets, self.lines):
            line_args = self.normalize_line_args(line_args)
            line_args[1] = np.array(line_args[1]) + offset
            handle = self.ax.plot(*line_args, **line_kwargs)
            self.handles.append(handle)

    @property
    def data_units_per_pixel(self):
        trans = self.ax.transData.transform
        inverse = ((trans((1, 1)) - trans((0, 0))) * self.ppd)
        return (1 / inverse[0], 1 / inverse[1])

    def normalize_line_args(self, args):
        def is_data_type(value):
            return isinstance(value, (
                np.array, np.ndarray, list, tuple
            ))

        assert is_data_type(args[0])

        if len(args) > 1 and is_data_type(args[1]) and len(args[0]) == len(args[1]):
            # looks like we have x and y data
            return args

        # otherwise we should pad the args with the x data
        return [range(len(args[0]))] + args

    def _resize(self, event=None):
        # Keep the trace in here until we can test appropriately.
        import pdb
        pdb.set_trace()
        """
        self.line.set_linewidth(lw)
        self.ax.figure.canvas.draw_idle()
        self.lw = lw
        """


def invisible_axes(ax):
    ax.grid(False)
    ax.set_axis_off()
    ax.patch.set_alpha(0)


def no_ticks(ax):
    ax.get_xaxis().set_ticks([])
    ax.get_yaxis().set_ticks([])



# pylint: disable=import-error

from PyQt5 import QtCore, QtWidgets

from .utils import PRETTY_KEYS, label, vlayout


class HelpDialog(QtWidgets.QDialog):
    def __init__(self, shortcuts=None):
        super().__init__()

        if shortcuts is None:
            shortcuts = []

        self.layout = QtWidgets.QVBoxLayout()

        keyboardShortcutsInfo = QtWidgets.QGroupBox(title='Keyboard Shortcuts')
        keyboardShortcutsLayout = QtWidgets.QGridLayout()
        for i, shortcut in enumerate(shortcuts):
            keyboardShortcutsLayout.addWidget(label(
                ', '.join(PRETTY_KEYS[k] for k in shortcut.chord), wordWrap=True), i, 0)
            keyboardShortcutsLayout.addWidget(label(shortcut.label), i, 1)

        keyboardShortcutsInfo.setLayout(keyboardShortcutsLayout)

        aboutInfo = QtWidgets.QGroupBox(title='About')
        aboutLayout = vlayout(
            label(
                'QtTool is the work of Conrad Stansbury, with much inspiration '
                'and thanks to the authors of ImageTool. QtTool is distributed '
                'as part of the PyARPES data analysis framework.',
                wordWrap=True
            ),
            label(
                'Complaints and feature requests should be directed to chstan@berkeley.edu.',
                wordWrap=True
            ),
        )
        aboutInfo.setLayout(aboutLayout)
        aboutInfo.setFixedHeight(150)

        self.layout.addWidget(keyboardShortcutsInfo)
        self.layout.addWidget(aboutInfo)
        self.setLayout(self.layout)

        self.setWindowTitle('QtTool - Help')
        self.setFixedSize(300, 500)

    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key_H or event.key() == QtCore.Qt.Key_Escape:
            self._main_window._help_dialog = None # pylint: disable=protected-access
            self.close()



from collections import namedtuple

import numpy as np

ParamType = namedtuple('ParamType', ['value', 'stderr'])


def param_getter(param_name, safe=True):
    """
    Constructs a function to extract a parameter value by name. Useful to extract data from inside an array of
    `lmfit.ModelResult` instances.

    :param param_name: Parameter name to retrieve. If you performed a composite model fit,
    make sure to include the prefix.
    :param safe: Guards against NaN values. This is typically desirable but sometimes it is advantageous
    to have NaNs fail an analysis quickly.
    :return:
    """
    if safe:
        safe_param = ParamType(value=np.nan, stderr=np.nan)

        def getter(x):
            try:
                return x.params.get(param_name, safe_param).value
            except:
                return np.nan

        return getter

    return lambda x: x.params[param_name].value


def param_stderr_getter(param_name, safe=True):
    """
    Constructs a function to extract a parameter value by name. Useful to extract data from inside an array of
    `lmfit.ModelResult` instances.

    :param param_name: Parameter name to retrieve. If you performed a composite model fit,
    make sure to include the prefix.
    :param safe: Guards against NaN values. This is typically desirable but sometimes it is advantageous
    to have NaNs fail an analysis quickly.
    :return:
    """

    if safe:
        safe_param = ParamType(value=np.nan, stderr=np.nan)

        def getter(x):
            try:
                return x.params.get(param_name, safe_param).stderr
            except:
                return np.nan

        return getter

    return lambda x: x.params[param_name].stderr



"""
This module is a work-in-progress. Ideally, we would one day like to offer
a simple graphical utility that can communicate with various ARPES DAQ hardware using
a file based interchange format. Currently, DAQ sequences roughly consistent with the
capabilities of the MAESTRO beamlines are supported via JSON.

More capabiities are also available for Tr-ARPES (shuffling to prevent laser and drift
skewing datasets) with more to come.
"""

import json
from itertools import chain, product
from pathlib import Path
from typing import Any, Iterable, Iterator, Union

import numpy as np

__all__ = ('JSONExperimentDriver', 'linspace', 'shuffled', 'move', 'comment', 'collect')


def flatten(lists):
    return chain.from_iterable([l if np.iterable(l) else [l] for l in lists])


class ExperimentTreeItem:
    def __add__(self, other):
        if isinstance(other, ExperimentTreeItem):
            return [self] + [other]

        return [self] + other

    def __mul__(self, other):
        return Product(self, other)


class Product(ExperimentTreeItem):
    items = None

    def __init__(self, *args):
        self.items = args
        self._iter = None

    def __len__(self):
        return np.product([len(item) for item in self.items])

    def __iter__(self):
        def safeiter(item_or_iterable: Union[Any, Iterable[Any]]) -> Iterator[Any]:
            try:
                return iter(item_or_iterable)
            except TypeError:
                return iter([item_or_iterable])

        self._iter = product(*[safeiter(i) for i in self.items])
        return self

    def __next__(self):
        return next(self._iter)

    def __repr__(self):
        front = '<Product>'
        content = '\n[\n{}\n]\n'.format((', \n'.join(['\t' + repr(f) for f in self.items])))

        back = '</ Product>'

        return front + content + back


class Collect(ExperimentTreeItem):
    def __init__(self, duration, configuration):
        self.duration = duration
        self.configuration = configuration

    def __repr__(self):
        return '<Collect duration={} configuration={} />'.format(self.duration, self.configuration)


class Move(ExperimentTreeItem):
    moveset = None
    wait_after = 0
    measure_while_moving = False
    backlash_compensate = False

    def __init__(self, wait_after=0, measure_while_moving=False, backlash_compensate=False, **kwargs):
        self.moveset = kwargs
        self.wait_after = wait_after
        self.measure_while_moving = measure_while_moving
        self.backlash_compensate = backlash_compensate

    def __repr__(self):
        return '<Move {} wait_after={} measure_while_moving={} backlash_compensate={}/>'.format(
            ' '.join(['{}={}'.format(k, v) for k, v in self.moveset.items()]),
            self.wait_after,
            self.measure_while_moving,
            self.backlash_compensate
        )


class Comment(ExperimentTreeItem):
    comment = ''

    def __init__(self, the_comment):
        self.comment = the_comment


class Shuffled(ExperimentTreeItem):
    values = None

    def __init__(self, values):
        self.values = values

    def __repr__(self):
        front = '<Shuffled>'
        try:
            content = '\n[\n{}\n]\n'.format((', \n'.join([
                '\t' + repr(v) for v in list(flatten(self.values))
            ])))
        except Exception: # pylint: disable=broad-except
            content = repr(self.values)

        back = '</ Shuffled>'
        return front + content + back


class Linspace(ExperimentTreeItem):
    start = 0
    stop = 0
    num = 0
    endpoint = True

    def __init__(self, f, start, stop, num, endpoint=True):
        self.f = f
        self.start = start
        self.stop = stop
        self.num = num
        self.endpoint = endpoint

    def __len__(self):
        return self.num

    def __iter__(self):
        return iter(self.values)

    @property
    def values(self):
        values = np.linspace(self.start, self.stop, self.num)
        return [self.f(v) for v in values]

    def __repr__(self):
        front = '<Linspace {} --> {}, {} steps endpoint={}>'.format(
            self.start, self.stop, self.num, self.endpoint)
        content = '\n[\n{}\n]\n'.format((', \n'.join(['\t' + repr(f) for f in self.values])))
        back = '</ Linspace>'
        return front + content + back

# pylint: disable=invalid-name

linspace = Linspace
shuffled = Shuffled
move = Move
comment = Comment
collect = Collect

# pylint: enable=invalid-name


class ExperimentDriver:
    queue_location = None
    seconds_per_frame = 1

    # units / second
    movement_speed = {
        'temp': 0.03, # deg kelvin per second

        'x': 1, # mm / sec
        'y': 1,
        'z': 1,

        'theta': 1 * np.pi / 180,
        'beta': 1 * np.pi / 180,
        'chi': 1 * np.pi / 180,
        'alpha': 0.1 * np.pi / 180,
        'psi': 100,
    }

    def __init__(self, queue_location=None):
        self.queue_location = queue_location

    @property
    def ext(self):
        return 'drive'

    def dump(self, file, input_object, **kwargs):
        file.write(self.dumps(input_object, **kwargs))

    def dumps(self, input_object, desired_total_time=None):
        return ''

    def dump_to_queue(self, name, input_object, **kwargs):
        if self.queue_location is None:
            raise ValueError('Must supply a queue location.')

        dump_path = Path(self.queue_location) / '{}.{}'.format(name, self.ext)
        with open(str(dump_path), 'w') as f:
            self.dump(f, input_object, **kwargs)

    def calculate_overhead(self, o):
        return self.calculate_duration(o)[1:]

    def calculate_duration(self, o):
        """
        Returns the anticipated scan duration, together with the
        fraction of time spent as movement or idle time overhead,
        the fraction of time spent in overhead in the spectrometer,
        and the fraction of acquisition time wasted due to overhead
        in sweeping (i.e. the DAQ "efficiency")

        I.e., if you received
        [600, 0.1, 0, 0.25]
        this means that the scan will take about 10 minutes, of which
        60 seconds (10%) of the time is spent in movement overhead, and
        no extra time is anticipated to be spent adjusting voltages
        in the spectrometer, and only 75% of the data recorded by
        the spectrometer will be kept.

        :param o:
        :return:
        """
        return [0, 0, 0, 0]


class ExperimentEncoder(json.JSONEncoder):
    def default(self, obj):  # pylint: disable=method-hidden
        if isinstance(obj, Product):
            return {
                'type': 'product',
                'items': obj.items,
            }
        if isinstance(obj, Linspace):
            return {
                'type': 'linspace',
                'start': obj.start,
                'stop': obj.stop,
                'num': obj.num,
                'endpoint': obj.endpoint,
                'values': obj.values,
            }
        if isinstance(obj, Move):
            return {
                'type': 'move',
                'wait_after': obj.wait_after,
                'measure_while_moving': obj.measure_while_moving,
                'backlash_compensate': obj.backlash_compensate,
                'moveset': obj.moveset
            }
        if isinstance(obj, Collect):
            return {
                'type': 'collect',
                'duration': obj.duration,
                'configuration': obj.configuration,
            }
        return super().default(obj)


class FlatExperimentEncoder(json.JSONEncoder):
    def default(self, obj):  # pylint: disable=method-hidden
        if isinstance(obj, Product):
            return flatten(iter(obj))
        if isinstance(obj, Linspace):
            return flatten(obj.values)
        if isinstance(obj, Move):
            return {
                'type': 'move',
                'wait_after': obj.wait_after,
                'measure_while_moving': obj.measure_while_moving,
                'backlash_compensate': obj.backlash_compensate,
                'moveset': obj.moveset
            }
        if isinstance(obj, Collect):
            return {
                'type': 'collect',
                'duration': obj.duration,
                'configuration': obj.configuration,
            }
        return super().default(obj)


class JSONExperimentDriver(ExperimentDriver):
    def dumps(self, o, desired_total_time=None):
        initial_pass = json.dumps({'sequence': o, 'configuration': {}}, cls=FlatExperimentEncoder)

        # we do an extra flattening pass here
        dict_values = json.loads(initial_pass)
        dict_values['sequence'] = flatten(dict_values['sequence'])
        return json.dumps(dict_values)



"""
Primitive support for loading Laue data from the NorthStar x-ray
backscattering DAQ program.
"""

# Laue file structure courtesy Jonathan Denlinger, MERLIN endstations at the ALS
# 16-bit binary Laue histogram (.hs2) file
# Format:  2 byte*256*256= 131072 long + header info at the end
# Northstar 6.0:  132028 bytes   (956 byte extra)
# Northstar 6.2.6.9:  134280 bytes  ( 3208 byte extra)
# header includes (offset from start of header):
#  byte 65536 / 131072 / 2364   = sample name (character string - read to double space)
#  byte 65587 / 131124 / 2416  = operator (character string - read to double space)
#  byte 65638 / 131176 / 2468  = date in mm/dd/yy format (8 character string)
#  byte 65811 / 131760 / 2984  = dwell time * 10 in seconds (word)
#  byte 65821 / 131776 / 3000  = mA
#  byte 65823 / 131780 / 3004  = kV
#  byte 131664 / 592 = index file name

import xarray
import typing
import numpy as np
from pathlib import Path

from arpes.provenance import provenance_from_file

northstar_62_69_dtype = np.dtype([
    ('pad1', 'B', (2364,),), # unused
    ('sample', 'S52'),
    ('user', 'S52'),
    ('comment', 'S512',),
    ('pad2', 'B', (228,),), # unused
])


def load_laue(path: typing.Union[Path, str]):
    if isinstance(path, str):
        path = Path(path)

    binary_data = path.read_bytes()
    table, header = binary_data[:131072], binary_data[131072:]

    table = np.fromstring(table, dtype=np.uint16).reshape(256, 256)
    header = np.fromstring(header, dtype=northstar_62_69_dtype).item()

    arr = xarray.DataArray(
        table,
        coords={'x': np.array(range(256)), 'y': np.array(range(256))},
        dims=['x', 'y',],
        attrs={
            'sample': header[1].split(b'\0')[0].decode('ascii'),
            'user': header[2].split(b'\0')[0].decode('ascii'),
            'comment': header[3].split(b'\0')[0].decode('ascii'),
        }
    )

    provenance_from_file(arr, str(path), {
        'what': 'Loaded Laue dataset from Northstar.',
        'by': 'load_laue',
    })

    return arr



"""
TODO: Standardize this module around support for some other library that has proper
Brillouin zone plotting, like in ASE.

This module also includes tools for masking regions of data against
Brillouin zones.
"""

import re
from collections import Counter, namedtuple

import matplotlib.path
import numpy as np

__all__ = ('bz_symmetry', 'bz_cutter', 'reduced_bz_selection',
           'reduced_bz_axes', 'reduced_bz_mask', 'reduced_bz_poly',
           'reduced_bz_axis_to', 'reduced_bz_E_mask', 'axis_along',
           'hex_cell', 'hex_cell_2d', 'orthorhombic_cell', 'process_kpath')


_SYMMETRY_TYPES = {
    ('G', 'X', 'Y'): 'rect',
    ('G', 'X'): 'square',
    ('G', 'X', 'BX'): 'hex',
}

_POINT_NAMES_FOR_SYMMETRY = {
    'rect': {'G', 'X', 'Y'},
    'square': {'G', 'X'},
    'hex': {'G', 'X', 'BX'},
}

SpecialPoint = namedtuple('SpecialPoint', ('name', 'negate', 'bz_coord'))


def parse_single_path(path):
    # first tokenize
    tokens = [name for name in re.split(r'([A-Z][a-z0-9]*(?:\([0-9,\s]+\))?)', path) if name]

    # normalize Gamma to G
    tokens = [token.replace('Gamma', 'G') for token in tokens]

    # convert to standard format
    points = []
    for token in tokens:
        name, rest = token[0], token[1:]
        negate = False
        if rest and rest[0] == 'n':
            negate = True
            rest = rest[1:]

        bz_coords = (0, 0, 0,)
        if rest:
            rest = ''.join(c for c in rest if c not in '( \t\n\r)')
            bz_coords = tuple([int(c) for c in rest.split(',')])

        if len(bz_coords) == 2:
            bz_coords = tuple(list(bz_coords) + [0])
        points.append(SpecialPoint(name=name, negate=negate, bz_coord=bz_coords))

    return points


def parse_path(paths):
    if isinstance(paths, str):

        # some manual string work in order to make sure we do not split on commas inside BZ indices
        idxs = []
        for i, p in enumerate(paths):
            if p == ',':
                c = Counter(paths[:i])
                if c['('] - c[')'] == 0:
                    idxs.append(i)

        paths = list(paths)
        for idx in idxs:
            paths[idx] = ':'

        paths = ''.join(paths)
        paths = paths.split(':')

    return [parse_single_path(p) for p in paths]


def special_point_to_vector(special_point, icell, special_points):
    base = np.dot(icell.T, special_points[special_point.name])

    if special_point.negate:
        base = -np.array(base)

    coord = np.array(special_point.bz_coord)
    return base + coord.dot(icell)


def process_kpath(paths, cell, special_points=None):
    promoted = False
    if len(cell) == 2:
        promoted = True
        cell = [c + [0] for c in cell] + [0, 0, 0]

    icell = np.linalg.inv(cell).T

    if special_points is None:
        from ase.dft.kpoints import get_special_points  # pylint: disable=import-error
        special_points = get_special_points(cell)

    points = [[special_point_to_vector(elem, icell, special_points) for elem in p]
               for p in parse_path(paths)]

    if promoted:
        print(points)
        pass

    return points

# Some common Brillouin zone formats
def orthorhombic_cell(a=1, b=1, c=1):
    return [[a, 0, 0], [0, b, 0], [0, 0, c]]


def hex_cell(a=1, c=1):
    return [[a, 0, 0], [-0.5 * a, 3 ** 0.5 / 2 * a, 0], [0, 0, c]]


def hex_cell_2d(a=1):
    return [[a, 0], [-0.5 * a, 3 ** 0.5 / 2 * a]]


def build_2dbz_poly(vertices=None, icell=None, cell=None):
    """
    Converts brillouin zone or equivalent information to a polygon mask
    that can be used to mask away data outside the zone boundary.
    :param vertices:
    :param icell:
    :param cell:
    :return:
    """
    from arpes.analysis.mask import raw_poly_to_mask
    from ase.dft.bz import bz_vertices  # pylint: disable=import-error

    assert(cell is not None or vertices is not None or icell is not None)

    if vertices is None:
        if icell is None:
            icell = np.linalg.inv(cell).T

        vertices = bz_vertices(icell)

    points, _ = vertices[0]  # points, normal
    points_2d = [p[:2] for p in points]

    return raw_poly_to_mask(points_2d)


def bz_symmetry(flat_symmetry_points):
    if isinstance(flat_symmetry_points, dict):
        flat_symmetry_points = flat_symmetry_points.items()

    largest_identified = 0
    symmetry = None

    point_names = set(k for k, _ in flat_symmetry_points)

    for points, sym in _SYMMETRY_TYPES.items():
        if all(p in point_names for p in points):
            if len(points) > largest_identified:
                symmetry = sym
                largest_identified = len(points)

    return symmetry

def reduced_bz_axis_to(data, S, include_E=False):
    symmetry = bz_symmetry(data.S.iter_own_symmetry_points)
    point_names = _POINT_NAMES_FOR_SYMMETRY[symmetry]

    symmetry_points, _ = data.S.symmetry_points()
    points = {k: v[0] for k, v in symmetry_points.items() if k in point_names}

    coords_by_point = {k: np.array([v.get(d, 0) for d in data.dims if d in v.keys() or include_E and d == 'eV'])
                       for k, v in points.items()}
    if symmetry == 'rect':
        if S == 'X':
            return coords_by_point['X'] - coords_by_point['G']
        return coords_by_point['Y'] - coords_by_point['G']
    elif symmetry == 'square':
        raise NotImplementedError()
        return coords_by_point['X'] - coords_by_point['G']
    elif symmetry == 'hex':
        if S == 'X':
            return coords_by_point['X'] - coords_by_point['G']
        return coords_by_point['BX'] - coords_by_point['G']
    else:
        raise NotImplementedError()


def reduced_bz_axes(data):
    symmetry = bz_symmetry(data.S.iter_own_symmetry_points)
    point_names = _POINT_NAMES_FOR_SYMMETRY[symmetry]

    symmetry_points, _ = data.S.symmetry_points()
    points = {k: v[0] for k, v in symmetry_points.items() if k in point_names}

    coords_by_point = {k: np.array([v[d] for d in data.dims if d in v.keys()]) for k, v in points.items()}
    if symmetry == 'rect':
        dx = coords_by_point['X'] - coords_by_point['G']
        dy = coords_by_point['Y'] - coords_by_point['G']
    elif symmetry == 'square':
        raise NotImplementedError()
        dx = coords_by_point['X'] - coords_by_point['G']
        dy = coords_by_point['X'] - coords_by_point['G']
    elif symmetry == 'hex':
        dx = coords_by_point['X'] - coords_by_point['G']
        dy = coords_by_point['BX'] - coords_by_point['G']
    else:
        raise NotImplementedError()

    return dx, dy


def axis_along(data, S):
    """
    Determines which axis lies principally along the direction G->S.
    :param data:
    :param S:
    :return:
    """
    symmetry = bz_symmetry(data.S.iter_own_symmetry_points)
    point_names = _POINT_NAMES_FOR_SYMMETRY[symmetry]

    symmetry_points, _ = data.S.symmetry_points()
    points = {k: v[0] for k, v in symmetry_points.items() if k in point_names}

    coords_by_point = {k: np.array([v[d] for d in data.dims if d in v.keys()]) for k, v in points.items()}

    dS = coords_by_point[S] - coords_by_point['G']

    max = -np.inf
    max_dim = None
    for dD, d in zip(dS, [d for d in data.dims if d != 'eV']):
        if np.abs(dD) > max:
            max = np.abs(dD)
            max_dim = d

    return max_dim


def reduced_bz_poly(data, scale_zone=False):
    symmetry = bz_symmetry(data.S.iter_own_symmetry_points)
    point_names = _POINT_NAMES_FOR_SYMMETRY[symmetry]
    dx, dy = reduced_bz_axes(data)
    if scale_zone:
        # should be good enough, reevaluate later
        dx = 3 * dx
        dy = 3 * dy

    symmetry_points, _ = data.S.symmetry_points()
    points = {k: v[0] for k, v in symmetry_points.items() if k in point_names}
    coords_by_point = {k: np.array([v.get(d, 0) for d in data.dims if d in v.keys()]) for k, v in points.items()}

    if symmetry == 'hex':
        return np.array([
            coords_by_point['G'],
            coords_by_point['G'] + dx,
            coords_by_point['G'] + dy,
        ])

    return np.array([
        coords_by_point['G'],
        coords_by_point['G'] + dx,
        coords_by_point['G'] + dx + dy,
        coords_by_point['G'] + dy,
    ])


def reduced_bz_E_mask(data, S, e_cut, scale_zone=False):
    symmetry_points, _ = data.S.symmetry_points()
    symmetry = bz_symmetry(data.S.iter_own_symmetry_points)
    point_names = _POINT_NAMES_FOR_SYMMETRY[symmetry]
    # bz_dims = tuple(d for d in data.dims if d in list(symmetry_points.values())[0][0].keys())


    symmetry_points, _ = data.S.symmetry_points()
    points = {k: v[0] for k, v in symmetry_points.items() if k in point_names}
    coords_by_point = {k: np.array([v.get(d,0) for d in data.dims if d in v.keys() or d == 'eV'])
                       for k, v in points.items()}

    dx_to = reduced_bz_axis_to(data, S, include_E=True)
    if scale_zone:
        dx_to = dx_to * 3
    dE = np.array([0 if d != 'eV' else e_cut for d in data.dims])

    poly_points = np.array([
        coords_by_point['G'],
        coords_by_point['G'] + dx_to,
        coords_by_point['G'] + dx_to + dE,
        coords_by_point['G'] + dE,
    ])

    skip_col = None
    for i in range(poly_points.shape[1]):
        if np.all(poly_points[:, i] == poly_points[0, i]):
            skip_col = i

    assert skip_col is not None
    selector_val = poly_points[0, skip_col]
    poly_points = np.concatenate((poly_points[:, 0:skip_col], poly_points[:, skip_col+1:]), axis=1)

    selector = dict()
    selector[data.dims[skip_col]] = selector_val
    sdata = data.sel(**selector, method='nearest')

    path = matplotlib.path.Path(poly_points)
    grid = np.array([a.ravel() for a in np.meshgrid(*[data.coords[d] for d in sdata.dims], indexing='ij')]).T
    mask = path.contains_points(grid)
    mask = np.reshape(mask, sdata.data.shape)
    return mask


def reduced_bz_mask(data, **kwargs):
    symmetry_points, _ = data.S.symmetry_points()
    bz_dims = tuple(d for d in data.dims if d in list(symmetry_points.values())[0][0].keys())

    poly_points = reduced_bz_poly(data, **kwargs)
    extra_dims_shape = tuple(len(data.coords[d]) for d in data.dims
                             if d in bz_dims)

    path = matplotlib.path.Path(poly_points)
    grid = np.array([a.ravel() for a in np.meshgrid(*[data.coords[d] for d in bz_dims], indexing='ij')]).T
    mask = path.contains_points(grid)
    mask = np.reshape(mask, extra_dims_shape)

    return mask


def reduced_bz_selection(data):
    mask = reduced_bz_mask(data)

    data = data.copy()
    data.data[np.logical_not(mask)] = np.nan

    return data


def bz_cutter(symmetry_points, reduced=True):
    """
    TODO UNFINISHED
    :param symmetry_points:
    :param reduced:
    :return:
    """
    def build_bz_mask(data):
        pass



    def cutter(data, cut_value=np.nan):
        mask = build_bz_mask(data)

        out = data.copy()
        out.data[mask] = cut_value

        return out

    return cutter



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io
import os

from setuptools import find_packages, setup
from setuptools.command.install import install

NAME = 'arpes'
DESCRIPTION = 'Modular data analysis code for angle resolved photoemission spectroscopy (ARPES)'
URL = 'https://gitlab.com/lanzara-group/python-arpes'
EMAIL = 'chstan@berkeley.edu'
AUTHOR = 'Conrad Stansbury'
REQUIRES_PYTHON = '>=3.5.0'

about = {}
with open('./arpes/__init__.py') as fp:
    exec(fp.read(), about)

VERSION = about['VERSION']

DEPENDENCY_GROUPS = {
    'core': [
        'tornado==4.5.3', # this version required due to Bokeh weirdness
        'astropy',

        # this version required as 0.10.1 introduces a change that forced all assignment
        # through Variable.get_compatible_data, which coerces to an array
        'xarray==0.9.6',

        'h5py==2.7.0',
        'netCDF4==1.3.0',
        'colorcet',

        'pint',
        'pandas',
        'dask',
        'numpy',
        'scipy',
        'lmfit>=0.9.13',
        'scikit-learn',

        # plotting
        'matplotlib>=3.0.3',
        'seaborn',
        'bokeh==0.12.10',
        'ipywidgets==7.0.1',

        # Misc deps
        'xlrd',
        'colorama',
        'titlecase',
        'openpyxl',
        'tqdm',
    ],
    'igor': ['igor==0.3.1'],
    'ml': [
        'scikit-learn',
        'scikit-image',
        'cvxpy',
        'libgcc',
    ],
}

requirements = [y for k, v in DEPENDENCY_GROUPS.items() for y in v if k not in {'igor', 'ml'}]

DEV_DEPENDENCIES = {
    'jupyter': [
        'jupyter',
        'ipython',
        'jupyter_contrib_nbextensions',
        'notebook>=5.7.0',
    ],
    'test': [
        'attrs==17.4.0',
        'pluggy==0.6.0',
        'py==1.5.2',
        'pytest==3.3.2',
        'setuptools==38.4.0',
    ]
}


with open('README.rst', 'r') as f_readme:
    long_description = f_readme.read()


DOCUMENTATION_URL = "https://stupefied-bhabha-ce8a9f.netlify.com/#/"

POST_INSTALL_MESSAGE = """
Documentation available at: {}

You should follow standard best practices for working with IPython and Jupyter.

To get the interactive volumetric data explorer `qt_tool` you will need to install
`PyQt5` and `pyqtgraph`. 

To use the Igor data loading libraries in PyARPES you will need to install the `igor` 
module from 'https://github.com/chstan/igorpy/tarball/712a4c4#egg=igor-0.3.1'.

Some functionality, including PCA/Factor Analysis decomposition tools, require 
additional heavy dependencies such as `scikit-learn` and `scikit-image`. 
""".format(DOCUMENTATION_URL)

packages = find_packages(exclude=('tests', 'source', 'info_session', 'docs', 'example_configuration',
                                  'conda', 'figures', 'exp', 'datasets', 'resources',))

here = os.path.abspath(os.path.dirname(__file__))

with io.open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    LONG_DESCRIPTION = '\n' + f.read()


class PostInstallCommand(install):
    """
    Provides some extra information and context after install.
    """

    def run(self):
        install.run(self)
        print(POST_INSTALL_MESSAGE)


# Where the magic happens:
setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,

    packages=packages,

    dependency_links=[
        'https://github.com/chstan/igorpy/tarball/712a4c4#egg=igor-0.3.1',
    ],
    install_requires=requirements,

    include_package_data=True,

    license='MIT',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Natural Language :: English',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows :: Windows 7',
        'Operating System :: Microsoft :: Windows :: Windows 8',
        'Operating System :: Microsoft :: Windows :: Windows 10',
        'Operating System :: Unix',
        'Operating System :: POSIX :: Linux',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    # $ setup.py publish support.
    cmdclass={
        'install': PostInstallCommand,
    },
)



import json
import os
import warnings
from abc import ABC, abstractmethod
from pathlib import Path

import numpy as np

import arpes.config
import colorcet as cc
import xarray as xr
from arpes.analysis.general import rebin
from arpes.io import load_dataset
from arpes.utilities import deep_equals
from typing import Union

__all__ = ('BokehInteractiveTool', 'CursorTool',)


class CursorTool:
    _cursor = None
    _cursor_info = None
    _horiz_cursor_x = None
    _horiz_cursor_y = None
    _vert_cursor_x = None
    _vert_cursor_y = None
    _cursor_lines = None
    _cursor_dims = None

    def __init__(self):
        self.data_range = {}
        self.arr = None
        self.app_context = {}

    @property
    def cursor_dims(self):
        return self._cursor_dims

    @cursor_dims.setter
    def cursor_dims(self, value):
        self._cursor_dims = value

    @property
    def cursor_dict(self):
        if self._cursor_dims is None:
            return None

        return dict(zip(self.cursor_dims, self.cursor))

    @property
    def cursor(self):
        return self._cursor

    def add_cursor_lines(self, figure):
        cursor_lines = figure.multi_line(xs=[self._horiz_cursor_x, self._vert_cursor_x],
                                         ys=[self._horiz_cursor_y, self._vert_cursor_y],
                                         line_color='white', line_width=2, line_dash='dotted')
        self._cursor_lines = cursor_lines
        return cursor_lines

    @cursor.setter
    def cursor(self, values):
        from bokeh.models.widgets.markups import Div

        if self._cursor_dims is None:
            try:
                self._cursor_dims = list(self.arr.dims)
            except AttributeError:
                pass

        self._cursor = values
        if self._cursor_info is None:
            self._cursor_info = Div(text='')
            self._horiz_cursor_x = list(self.data_range['x'])
            self._horiz_cursor_y = [0, 0]
            self._vert_cursor_x = [0, 0]
            self._vert_cursor_y = list(self.data_range['y'])
        else:
            self._vert_cursor_x[0] = self._vert_cursor_x[1] = self.cursor[0]
            self._horiz_cursor_y[0] = self._horiz_cursor_y[1] = self.cursor[1]

        self._cursor_info.text = '<h2>Cursor:</h2><span>({})</span>'.format(
            ', '.join("{0:.3f}".format(c) for c in self.cursor))

        if self._cursor_lines is not None:
            self._cursor_lines.data_source.data = {
                'xs': [self._horiz_cursor_x, self._vert_cursor_x],
                'ys': [self._horiz_cursor_y, self._vert_cursor_y],
            }

        try:
            self.app_context['cursor_dict'] = dict(zip(self._cursor_dims, self.cursor))
            #self.app_context['full_cursor'] =
        except AttributeError:
            pass


class BokehInteractiveTool(ABC):
    auto_rebin = True
    auto_zero_nans = True
    rebin_size = 800

    _debug_div = None

    @property
    def debug_div(self):
        from bokeh.models.widgets.markups import Div

        if self._debug_div is None:
            self._debug_div = Div(text='', width=300, height=100)

        return self._debug_div

    def __setattr__(self, name, value):
        if name == 'debug_text':
            self._debug_div.text = value

        super().__setattr__(name, value)

    def update_colormap_for(self, plot_name):
        def update_plot_colormap(attr, old, new):
            plot_data = self.plots[plot_name].data_source.data['image']
            low, high = np.min(plot_data), np.max(plot_data)
            dynamic_range = high - low
            self.color_maps[plot_name].update(low=low + new[0] / 100 * dynamic_range,
                                              high=low + new[1] / 100 * dynamic_range)

        return update_plot_colormap

    def init_bokeh_server(self):
        from bokeh.io import output_notebook

        if 'bokeh_configured' not in arpes.config.CONFIG:
            arpes.config.CONFIG['bokeh_configured'] = True

            # use a longer load_timeout for heavy tools
            output_notebook(hide_banner=True, load_timeout=10000)

            # Don't need to manually start a server in the manner of
            # https://matthewrocklin.com/blog//work/2017/06/28/simple-bokeh-server
            # according to
            # https://github.com/bokeh/bokeh/blob/0.12.10/examples/howto/server_embed/notebook_embed.ipynb

    def load_settings(self, **kwargs):
        self.settings = arpes.config.SETTINGS.get('interactive', {}).copy()
        for k, v in kwargs.items():
            if k not in self.settings:
                self.settings[k] = v

    @property
    def default_palette(self):
        from bokeh import palettes

        palette_options = {
            'viridis': palettes.viridis(256),
            'magma': palettes.magma(256),
            'coolwarm': cc.coolwarm,
        }

        return palette_options[self.settings.get('palette', 'viridis')]

    def __init__(self):
        self.settings = None
        self.app_context = {
            'data': None,
            'plots': {},
            'figures': {},
            'color_maps': {},
            'widgets': {},
        }

        self.init_bokeh_server()

    def __getattribute__(self, item):
        """
        Allow more convenient use of attributes from self.app_context. This is a bit strange.
        :param item:
        :return:
        """
        try:
            return super().__getattribute__(item)
        except AttributeError:
            if item in self.app_context:
                return self.app_context[item]

    @abstractmethod
    def tool_handler(self, doc):
        pass

    def make_tool(self, arr: Union[xr.DataArray, str], notebook_url=None,
                  notebook_handle=True, **kwargs):
        from bokeh.application import Application
        from bokeh.application.handlers import FunctionHandler
        from bokeh.io import show


        def generate_url(port):
            if port is None:
                return 'localhost:8888'

            return 'localhost:{}'.format(port)

        if notebook_url is None:
            if 'PORT' in arpes.config.CONFIG:
                notebook_url = 'localhost:{}'.format(arpes.config.CONFIG['PORT'])
            else:
                notebook_url = 'localhost:8888'

        if isinstance(arr, str):
            arr = load_dataset(arr)
            if 'cycle' in arr.dims and len(arr.dims) > 3:
                warnings.warn('Summing over cycle')
                arr = arr.sum('cycle', keep_attrs=True)

        if self.auto_zero_nans and {'kx', 'ky', 'kz', 'kp'}.intersection(set(arr.dims)):
            # We need to copy and make sure to clear any nan values, because bokeh
            # does not send these over the wire for some reason
            arr = arr.copy()
            np.nan_to_num(arr.values, copy=False)

        # rebin any axes that have more than 800 pixels
        if self.auto_rebin and np.any(np.asarray(arr.shape) > self.rebin_size):
            reduction = {d: (s // self.rebin_size) + 1 for d, s in arr.S.dshape.items()}
            warnings.warn('Rebinning with {}'.format(reduction))

            arr = rebin(arr, reduction=reduction)

            # TODO pass in a reference to the original copy of the array and make sure that
            # preparation tasks move over transparently

        self.arr = arr
        handler = FunctionHandler(self.tool_handler)
        app = Application(handler)
        show(app, notebook_url=notebook_url, notebook_handle=notebook_handle)

        return self.app_context


class SaveableTool(BokehInteractiveTool):
    def __init__(self, name=None):
        super().__init__()
        self.name = name
        self._last_save = None

    def make_tool(self, arr: Union[xr.DataArray, str], notebook_url=None,
                  notebook_handle=True, **kwargs):
        super().make_tool(arr, notebook_url=notebook_url, notebook_handle=notebook_handle, **kwargs)
        return self.app_context

    @property
    def filename(self):
        if self.name is None:
            return None

        return os.path.join(os.getcwd(), 'tools', 'tool-{}.json'.format(self.name))

    @property
    def path(self):
        return None if self.filename is None else Path(self.filename)

    def load_app(self):
        if self.name is None:
            return

        if not self.path.exists():
            return {}

        self.path.parent.mkdir(exist_ok=True)
        with open(self.filename, 'r') as f:
            self.deserialize(json.load(f))

    def save_app(self):
        if self.name is None:
            return

        data = self.serialize()
        if deep_equals(data, self._last_save):
            return

        self.path.parent.mkdir(exist_ok=True)
        with open(self.filename, 'w') as f:
            self._last_save = data
            json.dump(data, f)

    def deserialize(self, json_data):
        pass

    def serialize(self):
        return {}



import numpy as np
import xarray as xr

from arpes.provenance import update_provenance
from arpes.utilities.conversion.bounds_calculations import euler_to_kx, euler_to_ky, euler_to_kz

__all__ = ('convert_coordinates_to_kspace_forward',)


@update_provenance('Forward convert coordinates to momentum')
def convert_coordinates_to_kspace_forward(arr: xr.DataArray, **kwargs):
    """
    Forward converts all the individual coordinates of the data array
    :param arr:
    :param kwargs:
    :return:
    """

    arr = arr.copy(deep=True)

    skip = {'eV', 'cycle', 'delay', 'T'}
    keep = {'eV', }

    all = {k: v for k, v in arr.indexes.items() if k not in skip}
    kept = {k: v for k, v in arr.indexes.items() if k in keep}

    old_dims = list(all.keys())
    old_dims.sort()

    if not old_dims:
        return None

    dest_coords = {
        ('phi',): ['kp', 'kz'],
        ('theta',): ['kp', 'kz'],
        ('beta',): ['kp', 'kz'],
        ('phi', 'theta',): ['kx', 'ky', 'kz'],
        ('phi', 'beta',): ['kx', 'ky', 'kz'],
        ('hv', 'phi',): ['kx', 'ky', 'kz'],
        ('hv',): ['kp', 'kz'],
        ('hv', 'phi', 'beta'): ['kx', 'ky', 'kz'],
        ('hv', 'phi', 'theta'): ['kx', 'ky', 'kz'],
        ('hv', 'phi', 'psi'): ['kx', 'ky', 'kz'],
        ('hv', 'phi', 'chi'): ['kx', 'ky', 'kz'],
    }.get(tuple(old_dims))

    full_old_dims = old_dims + list(kept.keys())
    projection_vectors = np.ndarray(shape=tuple(len(arr.coords[d]) for d in full_old_dims), dtype=object)

    # these are a little special, depending on the scan type we might not have a phi coordinate
    # that aspect of this is broken for now, but we need not worry
    def broadcast_by_dim_location(data, target_shape, dim_location=None):
        if isinstance(data, xr.DataArray):
            if not data.dims:
                data = data.item()

        if isinstance(data, (int, float,)):
            return np.ones(target_shape) * data

        # else we are dealing with an actual array
        the_slice = [None] * len(target_shape)
        the_slice[dim_location] = slice(None, None, None)

        return np.asarray(data)[the_slice]

    raw_coords = {
        'phi': arr.coords['phi'].values - arr.S.phi_offset,
        'beta': (0 if arr.coords['beta'] is None else arr.coords['beta'].values) - arr.S.beta_offset,
        'theta': (0 if arr.coords['theta'] is None else arr.coords['theta'].values) - arr.S.theta_offset,
        'hv': arr.coords['hv'],
    }

    raw_coords = {k: broadcast_by_dim_location(v, projection_vectors.shape, full_old_dims.index(k) if k in full_old_dims else None)
                  for k, v in raw_coords.items()}

    # fill in the vectors
    binding_energy = broadcast_by_dim_location(arr.coords['eV'] - arr.S.work_function,
                                               projection_vectors.shape, full_old_dims.index('eV') if 'eV' in full_old_dims else None)
    photon_energy = broadcast_by_dim_location(arr.coords['hv'], projection_vectors.shape, full_old_dims.index('hv') if 'hv' in full_old_dims else None)
    kinetic_energy = binding_energy + photon_energy


    inner_potential = arr.S.inner_potential

    # some notes on angle conversion:
    # BL4 conventions
    # angle conventions are standard:
    # phi = analyzer acceptance
    # polar = perpendicular scan angle
    # theta = parallel to analyzer slit rotation angle

    # [ 1  0          0          ]   [  cos(polar) 0 sin(polar) ]   [ 0          ]
    # [ 0  cos(theta) sin(theta) ] * [  0          1 0          ] * [ k sin(phi) ]
    # [ 0 -sin(theta) cos(theta) ]   [ -sin(polar) 0 cos(polar) ]   [ k cos(phi) ]
    #
    # =
    #
    # [ 1  0          0          ]     [ sin(polar) * cos(phi) ]
    # [ 0  cos(theta) sin(theta) ] * k [ sin(phi) ]
    # [ 0 -sin(theta) cos(theta) ]     [ cos(polar) * cos(phi) ]
    #
    # =
    #
    # k ( sin(polar) * cos(phi),
    #     cos(theta)*sin(phi) + cos(polar) * cos(phi) * sin(theta),
    #     -sin(theta) * sin(phi) + cos(theta) * cos(polar) * cos(phi),
    #   )
    #
    # main chamber conventions, with no analyzer rotation (referred to as alpha angle in the Igor code
    # angle conventions are standard:
    # phi = analyzer acceptance
    # polar = perpendicular scan angle
    # theta = parallel to analyzer slit rotation angle

    # [ 1 0 0                    ]     [ sin(phi + theta) ]
    # [ 0 cos(polar) sin(polar)  ] * k [ 0                ]
    # [ 0 -sin(polar) cos(polar) ]     [ cos(phi + theta) ]
    #
    # =
    #
    # k (sin(phi + theta), cos(phi + theta) * sin(polar), cos(phi + theta) cos(polar), )
    #

    # for now we are setting the theta angle to zero, this only has an effect for vertical slit analyzers,
    # and then only when the tilt angle is very large

    # TODO check me
    raw_translated = {
        'kx': euler_to_kx(kinetic_energy, raw_coords['phi'], raw_coords['beta'], theta=0,
                          slit_is_vertical=arr.S.is_slit_vertical),
        'ky': euler_to_ky(kinetic_energy, raw_coords['phi'], raw_coords['beta'], theta=0,
                          slit_is_vertical=arr.S.is_slit_vertical),
        'kz': euler_to_kz(kinetic_energy, raw_coords['phi'], raw_coords['beta'], theta=0,
                          slit_is_vertical=arr.S.is_slit_vertical,
                          inner_potential=inner_potential),
    }

    if 'kp' in dest_coords:
        if np.sum(raw_translated['kx'] ** 2) > np.sum(raw_translated['ky'] ** 2):
            sign = raw_translated['kx'] / np.sqrt(raw_translated['kx'] ** 2 + 1e-8)
        else:
            sign = raw_translated['ky'] / np.sqrt(raw_translated['ky'] ** 2 + 1e-8)

        raw_translated['kp'] = np.sqrt(raw_translated['kx'] ** 2 + raw_translated['ky'] ** 2) * sign

    data_vars = {}
    for dest_coord in dest_coords:
        data_vars[dest_coord] = (full_old_dims, np.squeeze(raw_translated[dest_coord]))

    return xr.Dataset(data_vars, coords=arr.indexes)



import numpy as np
from bokeh import events

import xarray as xr
from arpes.analysis.band_analysis import fit_patterned_bands
from arpes.exceptions import AnalysisError
from arpes.models import band
from arpes.plotting.interactive_utils import CursorTool, SaveableTool
from arpes.utilities import normalize_to_spectrum

__all__ = ('BandTool',)


class BandTool(SaveableTool, CursorTool):
    """
    Two dimensional fitting band tool
    """
    auto_zero_nans = False
    auto_rebin = False

    def __init__(self, **kwargs):
        super().__init__(kwargs.pop('name', None))

        self.load_settings(**kwargs)

        self.app_main_size = self.settings.get('main_width', 600)
        self.app_marginal_size = self.settings.get('marginal_width', 300)
        self.active_band = None
        self.pointer_mode = 'band'

    def tool_handler(self, doc):
        from bokeh.layouts import row, column, widgetbox
        from bokeh.models.mappers import LinearColorMapper
        from bokeh.models import widgets
        from bokeh.plotting import figure

        if len(self.arr.shape) != 2:
            raise AnalysisError('Cannot use the band tool on non image-like spectra')

        arr = self.arr
        x_coords, y_coords = arr.coords[arr.dims[0]], arr.coords[arr.dims[1]]

        default_palette = self.default_palette

        self.app_context.update({
            'bands': {},
            'center_float': None,
            'data': arr,
            'data_range': {
                'x': (np.min(x_coords.values), np.max(x_coords.values)),
                'y': (np.min(y_coords.values), np.max(y_coords.values)),
            },
            'direction_normal': True,
            'fit_mode': 'mdc',
        })

        figures, plots, app_widgets = self.app_context['figures'], self.app_context['plots'], \
                                      self.app_context['widgets']
        self.cursor = [np.mean(self.data_range['x']),
                       np.mean(self.data_range['y'])]

        self.color_maps['main'] = LinearColorMapper(
            default_palette, low=np.min(arr.values), high=np.max(arr.values), nan_color='black')

        main_tools = ["wheel_zoom", "tap", "reset"]
        main_title = 'Band Tool: WARNING Unidentified'

        try:
            main_title = 'Band Tool: {}'.format(arr.S.label[:60])
        except:
            pass

        figures['main'] = figure(
            tools=main_tools, plot_width=self.app_main_size, plot_height=self.app_main_size, min_border=10,
            min_border_left=50,
            toolbar_location='left', x_axis_location='below', y_axis_location='right',
            title=main_title, x_range=self.data_range['x'],
            y_range=self.data_range['y'])
        figures['main'].xaxis.axis_label = arr.dims[0]
        figures['main'].yaxis.axis_label = arr.dims[1]
        figures['main'].toolbar.logo = None
        figures['main'].background_fill_color = "#fafafa"
        plots['main'] = figures['main'].image(
            [arr.values.T], x=self.app_context['data_range']['x'][0], y=self.app_context['data_range']['y'][0],
            dw=self.app_context['data_range']['x'][1] - self.app_context['data_range']['x'][0],
            dh=self.app_context['data_range']['y'][1] - self.app_context['data_range']['y'][0],
            color_mapper=self.app_context['color_maps']['main']
        )

        # add lines
        self.add_cursor_lines(figures['main'])
        band_lines = figures['main'].multi_line(xs=[], ys=[], line_color='white', line_width=1)

        def append_point_to_band():
            cursor = self.cursor
            if self.active_band in self.app_context['bands']:
                self.app_context['bands'][self.active_band]['points'].append(list(cursor))
                update_band_display()

        def click_main_image(event):
            self.cursor = [event.x, event.y]
            if self.pointer_mode == 'band':
                append_point_to_band()

        update_main_colormap = self.update_colormap_for('main')

        POINTER_MODES = [
            ('Cursor', 'cursor',),
            ('Band', 'band',),
        ]

        FIT_MODES = [
            ('EDC', 'edc',),
            ('MDC', 'mdc',),
        ]

        DIRECTIONS = [
            ('From Bottom/Left', 'forward',),
            ('From Top/Right', 'reverse'),
        ]

        BAND_TYPES = [
            ('Lorentzian', 'Lorentzian',),
            ('Voigt', 'Voigt',),
            ('Gaussian', 'Gaussian',)
        ]

        band_classes = {
            'Lorentzian': band.Band,
            'Gaussian': band.BackgroundBand,
            'Voigt': band.VoigtBand,
        }

        self.app_context['band_options'] = []

        def pack_bands():
            packed_bands = {}
            for band_name, band_description in self.app_context['bands'].items():
                if not band_description['points']:
                    raise AnalysisError('Band {} is empty.'.format(band_name))

                stray = None
                try:
                    stray = float(band_description['center_float'])
                except (KeyError, ValueError, TypeError):
                    try:
                        stray = float(self.app_context['center_float'])
                    except Exception:
                        pass

                packed_bands[band_name] = {
                    'name': band_name,
                    'band': band_classes.get(band_description['type'], band.Band),
                    'dims': self.arr.dims,
                    'params': {
                        'amplitude': {'min': 0},
                    },
                    'points': band_description['points'],
                }

                if stray is not None:
                    packed_bands[band_name]['params']['stray'] = stray

            return packed_bands

        def fit(override_data=None):
            packed_bands = pack_bands()
            dims = list(self.arr.dims)
            if 'eV' in dims:
                dims.remove('eV')
            angular_direction = dims[0]
            if isinstance(override_data, xr.Dataset):
                override_data = normalize_to_spectrum(override_data)
            return fit_patterned_bands(
                override_data if override_data is not None else self.arr, packed_bands,
                fit_direction='eV' if self.app_context['fit_mode'] == 'edc' else angular_direction,
                direction_normal=self.app_context['direction_normal']
            )

        self.app_context['pack_bands'] = pack_bands
        self.app_context['fit'] = fit

        self.pointer_dropdown = widgets.Dropdown(label='Pointer Mode', button_type='primary', menu=POINTER_MODES)
        self.direction_dropdown = widgets.Dropdown(label='Fit Direction', button_type='primary', menu=DIRECTIONS)
        self.band_dropdown = widgets.Dropdown(label='Active Band', button_type='primary',
                                              menu=self.app_context['band_options'])
        self.fit_mode_dropdown = widgets.Dropdown(label='Mode', button_type='primary', menu=FIT_MODES)
        self.band_type_dropdown = widgets.Dropdown(label='Band Type', button_type='primary', menu=BAND_TYPES)

        self.band_name_input = widgets.TextInput(placeholder='Band name...')
        self.center_float_widget = widgets.TextInput(placeholder='Center Constraint')
        self.center_float_copy = widgets.Button(label='Copy to all...')
        self.add_band_button = widgets.Button(label='Add Band')

        self.clear_band_button = widgets.Button(label='Clear Band')
        self.remove_band_button = widgets.Button(label='Remove Band')

        self.main_color_range_slider = widgets.RangeSlider(
            start=0, end=100, value=(0, 100,), title='Color Range')

        def add_band(band_name):
            if band_name not in self.app_context['bands']:
                self.app_context['band_options'].append((band_name, band_name,))
                self.band_dropdown.menu = self.app_context['band_options']
                self.app_context['bands'][band_name] = {
                    'type': 'Lorentzian',
                    'points': [],
                    'name': band_name,
                    'center_float': None,
                }

                if self.active_band is None:
                    self.active_band = band_name

                self.save_app()

        def on_copy_center_float():
            for band_name in self.app_context['bands'].keys():
                self.app_context['bands'][band_name]['center_float'] = self.app_context['center_float']
                self.save_app()

        def on_change_active_band(attr, old_band_id, band_id):
            self.app_context['active_band'] = band_id
            self.active_band = band_id

        def on_change_pointer_mode(attr, old_pointer_mode, pointer_mode):
            self.app_context['pointer_mode'] = pointer_mode
            self.pointer_mode = pointer_mode

        def set_center_float_value(attr, old_value, new_value):
            self.app_context['center_float'] = new_value
            if self.active_band in self.app_context['bands']:
                self.app_context['bands'][self.active_band]['center_float'] = new_value

            self.save_app()

        def set_fit_direction(attr, old_direction, new_direction):
            self.app_context['direction_normal'] = new_direction == 'forward'
            self.save_app()

        def set_fit_mode(attr, old_mode, new_mode):
            self.app_context['fit_mode'] = new_mode
            self.save_app()

        def set_band_type(attr, old_type, new_type):
            if self.active_band in self.app_context['bands']:
                self.app_context['bands'][self.active_band]['type'] = new_type

            self.save_app()

        def update_band_display():
            band_names = self.app_context['bands'].keys()
            band_lines.data_source.data = {
                'xs': [[p[0] for p in self.app_context['bands'][b]['points']] for b in band_names],
                'ys': [[p[1] for p in self.app_context['bands'][b]['points']] for b in band_names],
            }
            self.save_app()

        self.update_band_display = update_band_display

        def on_clear_band():
            if self.active_band in self.app_context['bands']:
                self.app_context['bands'][self.active_band]['points'] = []
                update_band_display()

        def on_remove_band():
            if self.active_band in self.app_context['bands']:
                del self.app_context['bands'][self.active_band]
                new_band_options = [b for b in self.app_context['band_options'] if b[0] != self.active_band]
                self.band_dropdown.menu = new_band_options
                self.app_context['band_options'] = new_band_options
                self.active_band = None
                update_band_display()

        # Attach callbacks
        self.main_color_range_slider.on_change('value', update_main_colormap)

        figures['main'].on_event(events.Tap, click_main_image)
        self.band_dropdown.on_change('value', on_change_active_band)
        self.pointer_dropdown.on_change('value', on_change_pointer_mode)
        self.add_band_button.on_click(lambda: add_band(self.band_name_input.value))
        self.clear_band_button.on_click(on_clear_band)
        self.remove_band_button.on_click(on_remove_band)
        self.center_float_copy.on_click(on_copy_center_float)
        self.center_float_widget.on_change('value', set_center_float_value)
        self.direction_dropdown.on_change('value', set_fit_direction)
        self.fit_mode_dropdown.on_change('value', set_fit_mode)
        self.band_type_dropdown.on_change('value', set_band_type)

        layout = row(column(figures['main']),
                     column(
                         widgetbox(
                             self.pointer_dropdown,
                             self.band_dropdown,
                             self.fit_mode_dropdown,
                             self.band_type_dropdown,
                             self.direction_dropdown,
                         ),
                         row(
                             self.band_name_input,
                             self.add_band_button,
                         ),
                         row(
                             self.clear_band_button,
                             self.remove_band_button,
                         ),
                         row(
                             self.center_float_widget,
                             self.center_float_copy
                         ),
                         widgetbox(
                             self._cursor_info,
                             self.main_color_range_slider,
                         )
                     ))

        doc.add_root(layout)
        doc.title = 'Band Tool'
        self.load_app()
        self.save_app()

    def serialize(self):
        return {
            'active_band': self.active_band,
            'band_options': self.band_options,
            'bands': self.bands,
            'center_float': self.app_context['center_float'],
            'cursor': self.cursor,
            'direction_normal': self.direction_normal,
            'fit_mode': self.fit_mode,
        }

    def deserialize(self, json_data):
        self.cursor = json_data.get('cursor', [0, 0])

        #self.active_band = json_data.get('active_band', None)

        self.app_context['center_float'] = json_data.get('center_float', None)
        self.app_context['bands'] = json_data.get('bands', {}) or {}
        self.app_context['direction_normal'] = json_data.get('direction_normal', True)
        self.app_context['fit_mode'] = json_data.get('fit_mode', 'mdc')
        self.app_context['band_options'] = json_data.get('band_options', [])

        self.center_float_widget.value = str(self.app_context['center_float'] or 0.1)
        self.band_dropdown.menu = self.app_context['band_options']

        self.update_band_display()



# pylint: disable=import-error

from PyQt5 import QtWidgets

from .utils import layout

__all__ = ('BinningInfoWidget',)

class BinningInfoWidget(QtWidgets.QGroupBox):
    def __init__(self, parent=None, root=None, axis_index=None):
        super().__init__(title=str(axis_index), parent=parent)
        self.root = root
        self.axis_index = axis_index

        self.spinbox = QtWidgets.QSpinBox()
        self.spinbox.setMaximum(2000)
        self.spinbox.setMinimum(1)
        self.spinbox.setValue(1)
        self.spinbox.valueChanged.connect(self.changeBinning)
        self.spinbox.editingFinished.connect(self.changeBinning)

        self.layout = layout(
            self.spinbox,
            widget=self,
        )

        self.recompute()

    def recompute(self):
        self.setTitle(self.root.data.dims[self.axis_index])

    def changeBinning(self):
        try:
            old_binning = self.root.binning
            old_binning[self.axis_index] = self.spinbox.value()
            self.root.binning = old_binning
        except:
            pass



# pylint: disable=import-error

import sys
import traceback

from PyQt5 import QtGui, QtCore, QtWidgets
import pyqtgraph as pg
import numpy as np
from collections import namedtuple, defaultdict

from arpes.utilities import normalize_to_spectrum
from arpes.typing import DataType
import arpes.config

from .utils import PRETTY_KEYS, pretty_key_event, KeyBinding, hlayout, vlayout, layout, tabs
from .HelpDialog import HelpDialog
from .AxisInfoWidget import AxisInfoWidget
from .DataArrayImageView import DataArrayImageView
from .BinningInfoWidget import BinningInfoWidget

__all__ = ('QtTool', 'qt_tool',)

ReactivePlotRecord = namedtuple('ReactivePlotRecord', ('dims', 'view', 'orientation',))

pg.setConfigOptions(antialias=True, foreground=(0, 0, 0), background=(255, 255, 255))


def clamp(x, low, high):
    if x <= low:
        return low
    if x >= high:
        return high
    return x


class CursorRegion(pg.LinearRegionItem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._region_width = 1
        self.lines[1].setMovable(False)

    def set_width(self, value):
        self._region_width = value
        self.lineMoved(0)

    def lineMoved(self, i):
        if self.blockLineSignal:
            return

        self.lines[1].setValue(self.lines[0].value() + self._region_width)
        self.prepareGeometryChange()
        self.sigRegionChanged.emit(self)


def patchedLinkedViewChanged(self, view, axis):
    """
    This still isn't quite right but it is much better than before. For some reason
    the screen coordinates of the PlotWidget are not being computed correctly, so
    we will just lock them as though they were perfectly aligned.

    This will clearly not work well for plots that have to be coordinated across
    different parts of the layout, but this will work for now.

    We also don't handle inverted axes for now.
    :param self:
    :param view:
    :param axis:
    :return:
    """
    if self.linksBlocked or view is None:
        return

    vr = view.viewRect()
    vg = view.screenGeometry()
    sg = self.screenGeometry()
    if vg is None or sg is None:
        return

    view.blockLink(True)

    try:
        if axis == pg.ViewBox.XAxis:
            upp = float(vr.width()) / vg.width()
            overlap = min(sg.right(), vg.right()) - max(sg.left(), vg.left())

            if overlap < min(vg.width() / 3, sg.width() / 3):
                x1 = vr.left()
                x2 = vr.right()
            else: # attempt to align
                x1 = vr.left()
                x2 = vr.right() + (sg.width() - vg.width()) * upp

            self.enableAutoRange(pg.ViewBox.XAxis, False)
            self.setXRange(x1, x2, padding=0)
        else:
            upp = float(vr.height()) / vg.height()
            overlap = min(sg.bottom(), vg.bottom()) - max(sg.top(), vg.top())

            if overlap < min(vg.height() / 3, sg.height() / 3):
                y1 = vr.top()
                y2 = vr.bottom()
            else: # again, attempt to align
                y1 = vr.top() # snap them at one side to the same coordinate

                # and scale the other side
                y2 = vr.bottom() + (sg.height() - vg.height()) * upp

            self.enableAutoRange(pg.ViewBox.YAxis, False)
            self.setYRange(y1, y2, padding=0)
    finally:
        view.blockLink(False)


pg.ViewBox.linkedViewChanged = patchedLinkedViewChanged


def patched_excepthook(exc_type, exc_value, exc_tb):
    enriched_tb = _add_missing_frames(exc_tb) if exc_tb else exc_tb
    traceback.print_exception(exc_type, exc_value, enriched_tb)

def _add_missing_frames(tb):
    result = fake_tb(tb.tb_frame, tb.tb_lasti, tb.tb_lineno, tb.tb_next)
    frame = tb.tb_frame.f_back
    while frame:
        result = fake_tb(frame, frame.f_lasti, frame.f_lineno, result)
        frame = frame.f_back
    return result

fake_tb = namedtuple('fake_tb', ('tb_frame', 'tb_lasti', 'tb_lineno', 'tb_next'))


class QtToolWindow(QtGui.QMainWindow, QtCore.QObject):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.tool = None
        self._help_dialog = None
        self._old_excepthook = sys.excepthook
        sys.excepthook = patched_excepthook

        self._keyBindings = [
            KeyBinding('Close', [QtCore.Qt.Key_Escape], self.do_close),
            KeyBinding('Toggle Help', [QtCore.Qt.Key_H], self.toggle_help),
            KeyBinding(
                'Scroll Cursor',
                [QtCore.Qt.Key_Left, QtCore.Qt.Key_Right, QtCore.Qt.Key_Up, QtCore.Qt.Key_Down],
                self.scroll
            ),
        ]

        QtGui.QGuiApplication.installEventFilter(self, self)

    def close(self):
        sys.excepthook = self._old_excepthook
        super(QtToolWindow, self).close()

    def do_close(self, event):
        self.close()

    def toggle_help(self, event):
        if self._help_dialog is None:
            self._help_dialog = HelpDialog(shortcuts=self._keyBindings)
            self._help_dialog.show()
            self._help_dialog._main_window = self # pylint: disable=protected-access
        else:
            self._help_dialog.close()
            self._help_dialog = None

    def window_print(self, *args, **kwargs):
        print(*args, **kwargs)

    def scroll(self, event):
        key_map = {
            QtCore.Qt.Key_Left: (0, -1),
            QtCore.Qt.Key_Right: (0, 1),
            QtCore.Qt.Key_Down: (1, -1),
            QtCore.Qt.Key_Up: (1, 1),
        }

        delta = key_map.get(event.key())
        if delta is not None and self.tool is not None:
            cursor = list(self.tool.context['cursor'])
            cursor[delta[0]] += delta[1]

            self.tool.update_cursor_position(cursor)

    def eventFilter(self, source, event):
        special_keys = [QtCore.Qt.Key_Down, QtCore.Qt.Key_Up, QtCore.Qt.Key_Left, QtCore.Qt.Key_Right]

        if event.type() in [QtCore.QEvent.KeyPress, QtCore.QEvent.ShortcutOverride]:
            if event.type() != QtCore.QEvent.ShortcutOverride or event.key() in special_keys:
                self.handleKeyPressEvent(event)

        return super().eventFilter(source, event)

    def handleKeyPressEvent(self, event):
        handled = False
        for binding in self._keyBindings:
            for combination in binding.chord:
                # only detect single keypresses for now
                if combination == event.key():
                    handled = True
                    binding.handler(event)

        if not handled:
            if arpes.config.SETTINGS.get('DEBUG', False):
                print(event.key())


class QtTool:
    """
    QtTool is an implementation of Image/Bokeh Tool based on PyQtGraph and PyQt5 for now we retain a number of the
    metaphors from BokehTool, including a "context" that stores the state, and can be used to programmatically interface
    with the tool
    """
    def __init__(self):
        self.settings = None
        self.context = {}
        self.data = None
        self.ninety_eight_percentile = None
        self.content_layout = None
        self.main_layout = None
        self.views = {}
        self.reactive_views = []
        self.registered_cursors = defaultdict(list)

        self.axis_info_widgets = []
        self.binning_info_widgets = []
        self.kspace_info_widgets = []

        self.axes_tab = None
        self.binning_tab = None
        self.kspace_tab = None
        self._binning = None

        self.settings = arpes.config.SETTINGS.copy()

    @property
    def binning(self):
        if self._binning is None:
            return [1 for _ in self.data.dims]

        return list(self._binning)

    @binning.setter
    def binning(self, value):
        different_binnings = [i for i, (nv, v) in enumerate(zip(value, self._binning)) if nv != v]
        self._binning = value

        for i in different_binnings:
            cursors = self.registered_cursors.get(i)
            for cursor in cursors:
                cursor.set_width(self._binning[i])

        self.update_cursor_position(self.context['cursor'], force=True)

    def print(self, *args, **kwargs):
        self.window.window_print(*args, **kwargs)

    def transpose(self, transpose_order):
        reindex_order = [self.data.dims.index(t) for t in transpose_order]
        self.data = self.data.transpose(*transpose_order)

        for widget in self.axis_info_widgets + self.binning_info_widgets:
            widget.recompute()

        self.update_cursor_position([self.context['cursor'][i] for i in reindex_order], force=True)

    def transpose_to_front(self, dim):
        if not isinstance(dim, str):
            dim = self.data.dims[dim]

        order = list(self.data.dims)
        order.remove(dim)
        order = [dim] + order
        self.transpose(order)

        # TODO update cursor, marginals, data

    def configure_image_widgets(self):
        if len(self.data.dims) == 2:
            self.generate_marginal_for((), 1, 0, 'xy', cursors=True)
            self.generate_marginal_for((1,), 0, 0, 'x', orientation='horiz')
            self.generate_marginal_for((0,), 1, 1, 'y', orientation='vert')

            self.views['xy'].view.setYLink(self.views['y'])
            self.views['xy'].view.setXLink(self.views['x'])

        if len(self.data.dims) == 3:
            self.generate_marginal_for((1, 2), 0, 0, 'x', orientation='horiz')
            self.generate_marginal_for((1,), 1, 0, 'xz')
            self.generate_marginal_for((2,), 2, 0, 'xy', cursors=True)
            self.generate_marginal_for((0, 1,), 0, 1, 'z', orientation='horiz', cursors=True)
            self.generate_marginal_for((0, 2,), 2, 2, 'y', orientation='vert')
            self.generate_marginal_for((0,), 2, 1, 'yz')

            self.views['xy'].view.setYLink(self.views['y'])
            self.views['xy'].view.setXLink(self.views['x'])
            self.views['xz'].view.setYLink(self.views['z'])
            self.views['xz'].view.setXLink(self.views['xy'].view)

        if len(self.data.dims) == 4:
            self.generate_marginal_for((1, 3), 0, 0, 'xz')
            self.generate_marginal_for((2, 3), 1, 0, 'xy', cursors=True)
            self.generate_marginal_for((0, 2,), 1, 1, 'yz')
            self.generate_marginal_for((0, 1,), 0, 1, 'zw', cursors=True)

    def set_colormap(self, colormap):
        sampled_colormap = colormap.colors[::51]
        cmap = pg.ColorMap(pos=np.linspace(0, 1, len(sampled_colormap)), color=sampled_colormap)
        for view_name, view in self.views.items():
            if isinstance(view, DataArrayImageView):
                view.setColorMap(cmap)

    def connect_cursor(self, dimension, the_line):
        self.registered_cursors[dimension].append(the_line)

        def connected_cursor(line: CursorRegion):
            new_cursor = list(self.context['cursor'])
            new_cursor[dimension] = line.getRegion()[0]
            self.update_cursor_position(new_cursor)

        the_line.sigRegionChanged.connect(connected_cursor)

    def update_cursor_position(self, new_cursor, force=False):
        old_cursor = list(self.context['cursor'])
        self.context['cursor'] = new_cursor

        def index_to_value(value, i):
            d = self.data.dims[i]
            c = self.data.coords[d].values
            return c[0] + value * (c[1] - c[0])

        self.context['value_cursor'] = [index_to_value(v, i) for i, v in enumerate(new_cursor)]

        changed_dimensions = [i for i, (x, y) in enumerate(zip(old_cursor, new_cursor)) if x != y]

        cursor_text = ','.join('{}: {:.4g}'.format(x, y)
                               for x, y in zip(self.data.dims, self.context['value_cursor']))
        self.window.statusBar().showMessage('({})'.format(cursor_text))

        # update axis info widgets
        for widget in self.axis_info_widgets + self.binning_info_widgets:
            widget.recompute()

        # update data
        def safe_slice(vlow, vhigh, axis=0):
            vlow, vhigh = int(min(vlow, vhigh)), int(max(vlow, vhigh))
            rng = len(self.data.coords[self.data.dims[axis]])
            vlow, vhigh = clamp(vlow, 0, rng), clamp(vhigh, 0, rng)
            
            if vlow == vhigh:
                vhigh = vlow + 1

            vlow, vhigh = clamp(vlow, 0, rng), clamp(vhigh, 0, rng)

            if vlow == vhigh:
                vlow = vhigh - 1

            return slice(vlow, vhigh)

        for reactive in self.reactive_views:
            if set(reactive.dims).intersection(set(changed_dimensions)) or force:
                try:
                    select_coord = dict(zip([self.data.dims[i] for i in reactive.dims],
                                            [safe_slice(int(new_cursor[i]), int(new_cursor[i] + self.binning[i]), i)
                                             for i in reactive.dims]))
                    if isinstance(reactive.view, DataArrayImageView):
                        image_data = self.data.isel(**select_coord)
                        if select_coord:
                            image_data = image_data.mean(list(select_coord.keys()))
                        reactive.view.setImage(image_data)

                    elif isinstance(reactive.view, pg.PlotWidget):
                        for_plot = self.data.isel(**select_coord)
                        if select_coord:
                            for_plot = for_plot.mean(list(select_coord.keys()))
                        for_plot = for_plot.values

                        cursors = [l for l in reactive.view.getPlotItem().items if isinstance(l, CursorRegion)]
                        reactive.view.clear()
                        for c in cursors:
                            reactive.view.addItem(c)

                        if reactive.orientation == 'horiz':
                            reactive.view.plot(for_plot)
                        else:
                            reactive.view.plot(for_plot, range(len(for_plot)))
                except IndexError:
                    pass

    def generate_marginal_for(self, dimensions, column, row, name=None, orientation='horiz', cursors=False):
        remaining_dims = [l for l in list(range(len(self.data.dims))) if l not in dimensions]
        if len(remaining_dims) == 1:
            widget = pg.PlotWidget(name=name)
            self.views[name] = widget
            if orientation == 'horiz':
                widget.setMaximumHeight(200)
            else:
                widget.setMaximumWidth(200)

            if cursors:
                cursor = CursorRegion(
                    orientation=CursorRegion.Vertical if orientation == 'vert' else CursorRegion.Horizontal,
                    movable=True)
                widget.addItem(cursor, ignoreBounds=False)
                self.connect_cursor(remaining_dims[0], cursor)
        else:
            assert len(remaining_dims) == 2
            widget = DataArrayImageView(name=name, root=self)
            widget.view.setAspectLocked(False)
            self.views[name] = widget

            widget.setHistogramRange(0, self.ninety_eight_percentile)
            widget.setLevels(0.05, 0.95)

            if cursors:
                cursor_vert = CursorRegion(orientation=CursorRegion.Vertical, movable=True)
                cursor_horiz = CursorRegion(orientation=CursorRegion.Horizontal, movable=True)
                widget.addItem(cursor_vert, ignoreBounds=True)
                widget.addItem(cursor_horiz, ignoreBounds=True)
                self.connect_cursor(remaining_dims[0], cursor_vert)
                self.connect_cursor(remaining_dims[1], cursor_horiz)

        self.reactive_views.append(
            ReactivePlotRecord(dims=dimensions, view=widget, orientation=orientation))
        self.content_layout.addWidget(widget, column, row)
        return widget

    @property
    def info_tab(self):
        return hlayout()

    def construct_axes_tab(self):
        inner_items = [AxisInfoWidget(axis_index=i, root=self) for i in range(len(self.data.dims))]
        return hlayout(*inner_items), inner_items

    def construct_binning_tab(self):
        binning_options = QtWidgets.QLabel('Options')
        inner_items = [BinningInfoWidget(axis_index=i, root=self) for i in range(len(self.data.dims))]

        return hlayout(binning_options, *inner_items), inner_items

    def construct_kspace_tab(self):
        inner_items = []
        return hlayout(*inner_items), inner_items

    def add_contextual_widgets(self):
        self.axes_tab, self.axis_info_widgets = self.construct_axes_tab()
        self.binning_tab, self.binning_info_widgets = self.construct_binning_tab()
        self.kspace_tab, self.kspace_info_widgets = self.construct_kspace_tab()

        self.tabs = tabs(
            ['Info', self.info_tab,],
            ['Axes', self.axes_tab,],
            ['Binning', self.binning_tab,],
            ['K-Space', self.kspace_tab,],
        )
        self.tabs.setFixedHeight(150)

        self.main_layout.addLayout(self.content_layout, 0, 0)
        self.main_layout.addWidget(self.tabs, 1, 0)

    def start(self):
        app = QtGui.QApplication([])

        win = QtToolWindow()
        win.resize(1100, 1100)
        win.setWindowTitle('QT Tool')
        cw = QtGui.QWidget()
        win.setCentralWidget(cw)
        self.window = win
        self.window.tool = self

        self.content_layout = QtGui.QGridLayout()
        self.main_layout = QtGui.QGridLayout()

        cw.setLayout(self.main_layout)

        # add main image widgets
        self.configure_image_widgets()
        self.add_contextual_widgets()
        import matplotlib.cm
        self.set_colormap(matplotlib.cm.viridis)
        win.show()

        # basic state initialization
        self.context.update({
            'cursor': [self.data.coords[d].mean().item() for d in self.data.dims],
        })

        ## Display the data
        self.update_cursor_position(self.context['cursor'], force=True)

        QtGui.QApplication.instance().exec()

    def set_data(self, data: DataType):
        data = normalize_to_spectrum(data)
        self.ninety_eight_percentile = np.percentile(data.values, (98,))[0]
        self.data = data
        self._binning = [1 for _ in self.data.dims]


def qt_tool(data: DataType):
    tool = QtTool()
    tool.set_data(data)
    tool.start()





import numpy as np

import xarray as xr
from arpes.provenance import update_provenance
from arpes.typing import DataType

__all__ = ('discretize_path', 'select_along_path', 'path_from_points',)


def path_from_points(data: DataType, symmetry_points_or_interpolation_points):
    """
    Acceepts a list of either tuples or point references. Point references can be string keys to `.attrs['symmetry_points']`
    This is the same behavior as `analysis.slice_along_path` and underlies the logic there.
    :param data:
    :param symmetry_points_or_interpolation_points:
    :return:
    """
    raise NotImplementedError('')


@update_provenance('Discretize Path')
def discretize_path(path: xr.Dataset, n_points=None, scaling=None):
    """
    Shares logic with slice_along_path
    :param path:
    :param n_points:
    :return:
    """

    if scaling is None:
        scaling = 1
    elif isinstance(scaling, xr.Dataset):
        scaling = {k: scaling[k].item() for k in scaling.data_vars}
    else:
        assert isinstance(scaling, dict)

    order = list(path.data_vars)
    if isinstance(scaling, dict):
        scaling = np.array(scaling[d] for d in order)

    def as_vec(ds):
        return np.array(list(ds[k].item() for k in order))

    def distance(a, b):
        return np.linalg.norm((as_vec(a) - as_vec(b)) * scaling)

    length = 0
    for idx_low, idx_high in (zip(path.index.values, path.index[1:].values)):
        coord_low, coord_high = path.sel(index=idx_low), path.sel(index=idx_high)
        length += distance(coord_low, coord_high)

    if n_points is None:
        # play with this until it seems reasonable
        n_points = int(length / 0.03)
    else:
        n_points = max(n_points - 1, 1)

    sep = length / n_points
    points = []
    distances = np.linspace(0, n_points - 1, n_points) * sep

    total_dist = 0
    for idx_low, idx_high in (zip(path.index.values, path.index[1:].values)):
        coord_low, coord_high = path.sel(index=idx_low), path.sel(index=idx_high)

        current_dist = distance(coord_low, coord_high)
        current_points = distances[distances < total_dist + current_dist]
        current_points = (current_points - total_dist) / current_dist
        distances = distances[len(current_points):]
        total_dist += current_dist

        points = points + list(np.outer(current_points, as_vec(coord_high) - as_vec(coord_low)) + as_vec(coord_low))

    points.append(as_vec(path.sel(index=path.index.values[-1])))

    new_index = np.array(range(len(points)))

    def to_dataarray(name):
        index = order.index(name)
        data = [p[index] for p in points]

        return xr.DataArray(np.array(data), {'index': new_index}, ['index'])

    return xr.Dataset({k: to_dataarray(k) for k in order})


@update_provenance('Select from data along a path')
def select_along_path(path: xr.Dataset, data: DataType, radius=None, n_points=None, fast=True, scaling=None, **kwargs):
    """
    Performs integration along a path. This functionally allows for performing a finite width
    cut (with finite width perpendicular to the local path direction) along some path,
    and integrating along this perpendicular selection. This allows for better statistics in
    oversampled data.

    :param path:
    :param data:
    :param radius: A number or dictionary of radii to use for the selection along different dimensions, if none is provided
    reasonable values will be chosen. Alternatively, you can pass radii via `{dim}_r` kwargs as well, i.e. 'eV_r' or 'kp_r'
    :param n_points: The number of points to interpolate along the path, by default we will infer a reasonable number
    from the radius parameter, if provided or inferred
    :param fast: If fast is true, will use rectangular selections rather than ellipsoid ones
    :return:
    """
    new_path = discretize_path(path, n_points, scaling)

    selections = []
    for _, view in new_path.T.iterate_axis('index'):
        selections.append(data.S.select_around(view, radius=radius, fast=fast, **kwargs))

    return xr.concat(selections, new_path.index)
